-----------------------------------------------------------------------------------------------
-- Client Lua Script for TargetView
-- Copyright Celess. Portions (C) NCsoft via TargetFrame. All rights reserved
-----------------------------------------------------------------------------------------------

require "Window"
require "Unit"
require "GameLib"
require "Apollo"
require "PathMission"
require "P2PTrading"


-----------------------------------------------------------------------------------------------
-- TargetView Module Definition
-----------------------------------------------------------------------------------------------
local _Version,_VersionMinor = 70923,0
local _AddonName,_VersionWindow = "TargetView", 3
local TargetView = {}

-----------------------------------------------------------------------------------------------
-- Constants
-----------------------------------------------------------------------------------------------

local _eSaveLevel = GameLib.CodeEnumAddonSaveLevel

local _defaultSettings = {		--type: 0 custom, 1 single check, 2 slider, 3 color, 4 radio, 5 text, 6 number, 7 check parts, 8 signal, 9 drop
	-- Normal
	-- display
		-- CheckVisibilityOptions
	["bShowClusters"] = { default=true, nControlType=1, strControlName="MainShowCluster", fnCallback="OnSettingRebuildFrames" },
	["bShowMyCluster"] = { default=true, nControlType=1, strControlName="MainShowClusterMine", fnCallback="OnSettingRebuildFrames" },
	["bShowTargetCluster"] = { default=true, nControlType=1, strControlName="MainShowClusterTarget", fnCallback="OnSettingRebuildFrames" },
	["bShowFocusCluster"] = { default=true, nControlType=1, strControlName="MainShowClusterFocus", fnCallback="OnSettingRebuildFrames" },
	["bShowMount"] = { default=true, nControlType=1, strControlName="MainShowMount", fnCallback="OnSettingClusterOption" },
	["bShowBot"] = { default=true, nControlType=1, strControlName="MainShowBot", fnCallback="OnSettingClusterOption" },
	["bShowPet"] = { default=true, nControlType=1, strControlName="MainShowPet", fnCallback="OnSettingClusterOption" },
	["bShowVehicle"] = { default=true, nControlType=1, strControlName="MainShowVehicle", fnCallback="OnSettingClusterOption" },
	["bShowOrganization"] = { default=true, nControlType=1, strControlName="MainShowOrganization", fnCallback="OnSettingClusterOption" },
	--["bShowGroup"] = { default=true, nControlType=1, strControlName="MainShowGroup" },

	-- Advanced
	-- style
	["bFontLargeName"] = { default=false, nControlType=1, strControlName="AdvFontNameLarge", fnCallback="OnSettingFontChanged" },
	["bColorLightEnemy"] = { default=false, nControlType=1, strControlName="AdvColorEnemyLight", fnCallback="OnSettingFontChanged" },

	-- Other

	sSettingsName = "Default",
	nSettingsVersion = 1,
	tSettingsCustom = {	},
}

local _resetSettings = {
	{
		sSettingsName = "Default",
		sSettingsTooltip = "Default settings. These are the same as the current Carbine settings for Nameplates.",
	}, {
		sSettingsName = "Updated",
		sSettingsTooltip = "A more progressive style than Default, with a few of the newer TargetView features and refinements. Changes: Larger Name Font enabled, Lighter Enemy Color enabled.",
		bFontLargeName = true,
		bColorLightEnemy = true,
	}, {
		sSettingsName = "No Clusters",
		sSettingsTooltip = "Disable mini-targets completely. Changes: Larger Name Font enabled, Lighter Enemy Color enabled, Mini-Target Clusters disabled.",
		bShowClusters = false,
		bFontLargeName = true,
		bColorLightEnemy = true,
	}, {
		sSettingsName = "No Pets",
		sSettingsTooltip = "Disable pet mini-targets. Changes: Larger Name Font enabled, Lighter Enemy Color enabled, PETS disabled.",
		bShowPet = false,
		bFontLargeName = true,
		bColorLightEnemy = true,
	}, {
		sSettingsName = "Custom",
		sSettingsTooltip = "Custom defaults. Use the Save button at any time to save current settings to this Custom preset. Load them again by choosing this Custom preset from the list.",
	},
}
for _,v in pairs(_resetSettings) do
	_resetSettings[v.sSettingsName:lower()] = v
end

local _eDisposition = Unit.CodeEnumDisposition
local _eStateVulnerability = Unit.CodeEnumCCState.Vulnerability
local _eRewardInfoType = Unit.CodeEnumRewardInfoType
local _eRank = Unit.CodeEnumRank
local _eClass = GameLib.CodeEnumClass

local _eTintDisposition = {
	Hostile = _eDisposition.Hostile,
	Neutral = _eDisposition.Neutral,
	Friendly = _eDisposition.Friendly,
	Unknown = _eDisposition.Unknown,
	Vulnerable = _eDisposition.Unknown + 1,
}

local _eTintType = {
	Critical = 1,
	Warning = 2,
	Good = 3,
	Unknown = 4,
	Vulnerable = 5,
}

local _nHealthRed					= 0.3
local _nHealthYellow				= 0.5
--local _nHealthGreen
local _nHealthGlow					= 0.96


local _tHealthBar = {
	[_eTintType.Critical] = "spr_TargetFrame_HealthFillRed",
	[_eTintType.Warning] = "spr_TargetFrame_HealthFillYellow",
	[_eTintType.Good] = "spr_TargetFrame_HealthFillGreen",
	[_eTintType.Unknown] = "spr_TargetFrame_HealthFillYellow",
	[_eTintType.Vulnerable] = "spr_TargetFrame_HealthFillVulernable",
}

local _tHealthBarFlipped = {
	[_eTintType.Critical] = "spr_TargetFrame_HealthFillRedFlipped",
	[_eTintType.Warning] = "spr_TargetFrame_HealthFillYellowFlipped",
	[_eTintType.Good] = "spr_TargetFrame_HealthFillGreenFlipped",
	[_eTintType.Unknown] = "spr_TargetFrame_HealthFillYellowFlipped",
	[_eTintType.Vulnerable] = "spr_TargetFrame_HealthFillVulernableFlipped",
}

local _tClusterBar = {
	[_eTintType.Critical] = "spr_TargetFrame_ClusterHealthRed",
	[_eTintType.Warning] = "spr_TargetFrame_ClusterHealthYellow",
	[_eTintType.Good] = "spr_TargetFrame_ClusterHealthGreen",
	[_eTintType.Unknown] = "spr_TargetFrame_ClusterHealthYellow",
	[_eTintType.Vulnerable] = "spr_TargetFrame_ClusterHealthRed",
}

local _eVisiText = {
	Mouseover = 1,
	ShowXY = 2,
	ShowPercent = 3,
	Unknown = 4,
}

local _nMaxLevel					= 50
local _nFrameOfsBoth				= 0
local _nFrameOfsOne					= 27
local _nFrameOfsNone				= 61

local _tTextColors = {
	[_eDisposition.Hostile]  = ApolloColor.new("DispositionHostile"),
	[_eDisposition.Neutral]  = ApolloColor.new("DispositionNeutral"),
	[_eDisposition.Friendly] = ApolloColor.new("DispositionFriendly"),
	[_eDisposition.Unknown]  = ApolloColor.new("DispositionNeutral"),
}

local _tToTTooltipText = {
	[_eDisposition.Hostile]  = Apollo.GetString("TargetFrame_HostileTooltip"),
	[_eDisposition.Neutral]  = Apollo.GetString("TargetFrame_NeutralTooltip"),
	[_eDisposition.Friendly] = Apollo.GetString("TargetFrame_FriendlyTooltip"),
	[_eDisposition.Unknown]  = Apollo.GetString("TargetFrame_NeutralTooltip"),
}

local _tToTBackerSprite = {
	[_eDisposition.Hostile]  = "spr_TargetFrame_ToTHostile",
	[_eDisposition.Neutral]  = "spr_TargetFrame_ToTNeutral",
	[_eDisposition.Friendly] = "spr_TargetFrame_ToTFriendly",
	[_eDisposition.Unknown]  = "spr_TargetFrame_ToTNeutral",
}

local _tToTDispositionColor = {
	[_eDisposition.Hostile]  = ApolloColor.new("ffff6d66"),
	[_eDisposition.Neutral]  = ApolloColor.new("fffaff66"),
	[_eDisposition.Friendly] = ApolloColor.new("ff00f7de"),
	[_eDisposition.Unknown]  = ApolloColor.new("fffaff66"),
}

local _tDispositionSpriteLabel = {
	[_eDisposition.Hostile]  = "Hostile",
	[_eDisposition.Neutral]  = "Neutral",
	[_eDisposition.Friendly] = "Friendly",
	[_eDisposition.Unknown]  = "Hostile",
}

local _tRareSprite = {
	[_eDisposition.Hostile]  = "spr_TargetFrame_FrameHostileRare",
	[_eDisposition.Neutral]  = "spr_TargetFrame_FrameNeutralRare",
	[_eDisposition.Friendly] = "spr_TargetFrame_FrameFriendlyRare",
	[_eDisposition.Unknown]  = "spr_TargetFrame_FrameNeutralRare",		-- Hostile
}
local _tRareSpriteFlipped = {
	[_eDisposition.Hostile]  = "spr_TargetFrame_FrameHostileRareFlipped",
	[_eDisposition.Neutral]  = "spr_TargetFrame_FrameNeutralRareFlipped",
	[_eDisposition.Friendly] = "spr_TargetFrame_FrameFriendlyRareFlipped",
	[_eDisposition.Unknown]  = "spr_TargetFrame_FrameNeutralRareFlipped",
}

local _tEliteSprite = {
	[_eDisposition.Hostile]  = "spr_TargetFrame_FrameHostileElite",
	[_eDisposition.Neutral]  = "spr_TargetFrame_FrameNeutralElite",
	[_eDisposition.Friendly] = "spr_TargetFrame_FrameFriendlyElite",
	[_eDisposition.Unknown]  = "spr_TargetFrame_FrameNeutralElite",		-- Hostile
}
local _tEliteSpriteFlipped = {
	[_eDisposition.Hostile]  = "spr_TargetFrame_FrameHostileEliteFlipped",
	[_eDisposition.Neutral]  = "spr_TargetFrame_FrameNeutralEliteFlipped",
	[_eDisposition.Friendly] = "spr_TargetFrame_FrameFriendlyEliteFlipped",
	[_eDisposition.Unknown]  = "spr_TargetFrame_FrameNeutralEliteFlipped",
}

local _tSplitSprite = {
	[_eDisposition.Hostile]  = "spr_TargetFrame_HealthSplitHostile",
	[_eDisposition.Neutral]  = "spr_TargetFrame_HealthSplitNeutral",
	[_eDisposition.Friendly] = "spr_TargetFrame_HealthSplitFriendly",
	[_eDisposition.Unknown]  = "spr_TargetFrame_HealthSplitNeutral",		-- Hostile
}
local _tSplitSpriteFlipped = {
	[_eDisposition.Hostile]  = "spr_TargetFrame_HealthSplitHostileFlipped",
	[_eDisposition.Neutral]  = "spr_TargetFrame_HealthSplitNeutralFlipped",
	[_eDisposition.Friendly] = "spr_TargetFrame_HealthSplitFriendlyFlipped",
	[_eDisposition.Unknown]  = "spr_TargetFrame_HealthSplitNeutralFlipped",		-- Hostile
}

local _eBackerState = {
	Tagged = _eDisposition.Unknown + 1,
}
local _tBackerSprite = {
	[_eDisposition.Hostile]  = "spr_TargetFrame_FrameHostile",
	[_eDisposition.Neutral]  = "spr_TargetFrame_FrameNeutral",
	[_eDisposition.Friendly] = "spr_TargetFrame_FrameFriendly",
	[_eDisposition.Unknown]  = "spr_TargetFrame_FrameNeutral",		-- Hostile
	[_eBackerState.Tagged]   = "spr_TargetFrame_FrameTapped",
}
local _tBackerSpriteFlipped = {
	[_eDisposition.Hostile]  = "spr_TargetFrame_FrameHostileFlipped",
	[_eDisposition.Neutral]  = "spr_TargetFrame_FrameNeutralFlipped",
	[_eDisposition.Friendly] = "spr_TargetFrame_FrameFriendlyFlipped",
	[_eDisposition.Unknown]  = "spr_TargetFrame_FrameNeutralFlipped",		-- Hostile
	[_eBackerState.Tagged]   = "spr_TargetFrame_FrameTappedFlipped",
}

local _tFactionIdToType = {										-- model frame adornment, attachment
	[Unit.CodeEnumFaction.ExilesPlayer]		= "Exile",
	[171]									= "Exile",			-- exile npc
	[Unit.CodeEnumFaction.DominionPlayer]	= "Dominion",
	[170]									= "Dominion",		-- dominion npc
}

local _eAttachment = {
	None = 0,
	Vehicle = 1,
	Iconic = 2,
}
local _tAttachmentSprite = {
	[1] = "spr_TargetFrame_FrameVehicle",
	[2] = "spr_TargetFrame_FrameFriendlyIconicDominion",
	[3] = "spr_TargetFrame_FrameFriendlyIconicExile",
	[4] = "spr_TargetFrame_FrameHostileIconicDominion",
	[5] = "spr_TargetFrame_FrameHostileIconicExile"
}
local _tAttachmentSpriteFlipped = {
	[1] = "spr_TargetFrame_FrameVehicleFlipped",
	[2] = "spr_TargetFrame_FrameFriendlyIconicDominionFlipped",
	[3] = "spr_TargetFrame_FrameFriendlyIconicExileFlipped",
	[4] = "spr_TargetFrame_FrameHostileIconicDominionFlipped",
	[5] = "spr_TargetFrame_FrameHostileIconicExileFlipped"
}

local _eCastState = {
	Value = 1,
	Locked = 2,
	MoO	= 3,
	Unknown = 4,			-- unknown or dont care
}

local _tInteruptArmorSprite = {
	[_eCastState.Value] = "spr_TargetFrame_InterruptArmor_Value",
	[_eCastState.Locked] = "spr_TargetFrame_InterruptArmor_Infinite",
	[_eCastState.MoO] = "spr_TargetFrame_InterruptArmor_MoO",
	[_eCastState.Unknown] = "",
}

local _tCastBarSprite = {
	[_eCastState.Value] = "spr_TargetFrame_FrameCastBar",
	[_eCastState.Locked] = "spr_TargetFrame_FrameCastBarLocked",
	[_eCastState.MoO] = "spr_TargetFrame_FrameCastBarVulnerable",
	[_eCastState.Unknown] = "",
}
local _tCastBarSpriteFlipped = {
	[_eCastState.Value] = "spr_TargetFrame_FrameCastBarFlipped",
	[_eCastState.Locked] = "spr_TargetFrame_FrameCastBarLockedFlipped",
	[_eCastState.MoO] = "spr_TargetFrame_FrameCastBarVulnerableFlipped",
	[_eCastState.Unknown] = "",
}

local _tCastBarFillSprite = {
	[_eCastState.Value] = "spr_TargetFrame_FrameCastBarFill",
	[_eCastState.Locked] = "spr_TargetFrame_FrameCastBarFillLocked",
	[_eCastState.MoO] = "spr_TargetFrame_FrameCastBarFillVulnerable",
	[_eCastState.Unknown] = "",
}
local _tCastBarFillSpriteFlipped = {
	[_eCastState.Value] = "spr_TargetFrame_FrameCastBarFillFlipped",
	[_eCastState.Locked] = "spr_TargetFrame_FrameCastBarFillLockedFlipped",
	[_eCastState.MoO] = "spr_TargetFrame_FrameCastBarFillVulnerableFlipped",
	[_eCastState.Unknown] = "",
}

local _eCastType = Unit.CodeEnumCastBarType

local _tCastIconSprite = {
	[_eCastType.None] = "",
	[_eCastType.Normal] = "",
	[_eCastType.Telegraph_Backlash] = "CRB_TargetFrameSprites:sprTF_CastIconInterrupt",
	[_eCastType.Telegraph_Evade] = "CRB_TargetFrameSprites:sprTF_CastIconEvade",
}

local _ePlateType = {
	unitPlayer = 1,
	unitTarget = 2,
	unitFocus = 3,
}

local _tCarbineOptionFlip = {
	[_ePlateType.unitPlayer] = "bMyUnitFrameFlipped",
	[_ePlateType.unitTarget] = "bTargetFrameFlipped",
	[_ePlateType.unitFocus] = "bFocusFrameFlipped",
}

local _tDefaultPosition = {
	[_ePlateType.unitPlayer] = { fPoints = { 0.24, 1, 0.24, 1 }, nOffsets = { 0, -324, 358, -220 } },
	[_ePlateType.unitTarget] = { fPoints = { 0.76, 1, 0.76, 1 }, nOffsets = { -358, -324, 0, -220 } },
	[_ePlateType.unitFocus] = { fPoints = { 0, 0.5, 0, 0.5 }, nOffsets = { 30, -22, 388, 82 } },
}

local _tWindowManagementLabel= {
	[_ePlateType.unitPlayer] = Apollo.GetString("OptionsHUD_MyUnitFrameLabel"),
	[_ePlateType.unitTarget] = Apollo.GetString("OptionsHUD_TargetFrameLabel"),
	[_ePlateType.unitFocus] =  Apollo.GetString("OptionsHUD_FocusTargetLabel"),
}

local _tMarkerToSprite = {
	[1] = "Icon_Windows_UI_CRB_Marker_Bomb",
	[2] = "Icon_Windows_UI_CRB_Marker_Ghost",
	[3] = "Icon_Windows_UI_CRB_Marker_Mask",
	[4] = "Icon_Windows_UI_CRB_Marker_Octopus",
	[5] = "Icon_Windows_UI_CRB_Marker_Pig",
	[6] = "Icon_Windows_UI_CRB_Marker_Chicken",
	[7] = "Icon_Windows_UI_CRB_Marker_Toaster",
	[8] = "Icon_Windows_UI_CRB_Marker_UFO",
}

--TODO: break these out onto options
local _crGroup				= ApolloColor.new("BubbleTextShout")
--local _crGuild_Dark		= ApolloColor.new("DispositionGuildmateUnflagged")
--local _crGuild_Light		= ApolloColor.new("FFB970DB")	--FFAF5CD6 10% -- lightened about 20%. the regular purple color is very hard to read as its very dark against the black outline
--local _crGuild			= _tTextColors[_eDisposition.Friendly]
--local _crWarparty			= _tTextColors[_eDisposition.Friendly]
local _crAllyUnflag			= _tTextColors[_eDisposition.Friendly]
local _crAlly				= _tTextColors[_eDisposition.Friendly]
local _crEnemyUnflag 		= ApolloColor.new("UI_TextHoloBody")
local _crEnemy_Dark 		= _tTextColors[_eDisposition.Hostile]
local _crEnemy_Light 		= ApolloColor.new("FFFFF322A") --FA3A26") --f7211d")	--FA2521")		--FFFF211C") --FFFF3633" 20% --FFFF1D1A 10% ffff0500-- lightened about 20%. the regular red color is very hard to read as its very dark against the black outline
local _crEnemy				= _crEnemy_Dark
local _crEnemyMismatch		= _tTextColors[_eDisposition.Neutral]

local _sFont_Default	= "Nameplates"			-- CRB_Header10_O is "Nameplates"
local _sFont_Large		= "CRB_Interface10_BO"
local _sFont_Larger		= "CRB_Interface10_BO"
local _sFont_Large85	= "CRB_Interface11_BO"
local _sFont_Larger85	= "CRB_Interface11_BO"
local _sFont_Large75	= "CRB_Interface12_BO"
local _sFont_Larger75	= "CRB_Interface12_BO"
local _sFont_Large70	= "CRB_Header12_O"		-- panic values. very very common, people often use 1280 and .75 ui scale to play the game at all
local _sFont_Larger70	= "CRB_Header13_O"
local _sFont			= _sFont_Default
--local _sGuildFont			= "CRB_Interface9_BO"

local _csScalingHex		= "ffffbf80"
local _crScaling 		= CColor.new(1.0, 191/255, 128/255, 0.7)		--  "ffffbf80"

local _nConOffset = 5
local _nConMin = 1
local _nConMax = 9

local _arConColor = {						-- level differential color
	[1] = ApolloColor.new("ConTrivial"),	-- -4
	[2] = ApolloColor.new("ConInferior"),	-- -3
	[3] = ApolloColor.new("ConMinor"),
	[4] = ApolloColor.new("ConEasy"),
	[5] = ApolloColor.new("ConAverage"),
	[6] = ApolloColor.new("ConModerate"),
	[7] = ApolloColor.new("ConTough"),
	[8] = ApolloColor.new("ConHard"),
	[9] = ApolloColor.new("ConImpossible"),
}

local _arConTitleColor = {
	[1] = "ff7d7d7d",
	[2] = "ff01ff07",
	[3] = "ff01fcff",
	[4] = "ff597cff",
	[5] = "ffffffff",
	[6] = "ffffff00",
	[7] = "ffff8000",
	[8] = "ffff0000",
	[9] = "ffff00ff",
}

local _arConTitle = {
	[1] = Apollo.GetString("TargetFrame_Trivial"),
	[2] = Apollo.GetString("TargetFrame_Inferior"),
	[3] = Apollo.GetString("TargetFrame_Minor"),
	[4] = Apollo.GetString("TargetFrame_Easy"),
	[5] = Apollo.GetString("TargetFrame_Average"),
	[6] = Apollo.GetString("TargetFrame_Moderate"),
	[7] = Apollo.GetString("TargetFrame_Tough"),
	[8] = Apollo.GetString("TargetFrame_Hard"),
	[9] = Apollo.GetString("TargetFrame_Impossible"),
}

local _arConBody = {
	[1] = Apollo.GetString("TargetFrame_NoXP"),
	[2] = Apollo.GetString("TargetFrame_VeryReducedXP"),
	[3] = Apollo.GetString("TargetFrame_ReducedXP"),
	[4] = Apollo.GetString("TargetFrame_SlightlyReducedXP"),
	[5] = Apollo.GetString("TargetFrame_StandardXP"),
	[6] = Apollo.GetString("TargetFrame_SlightlyMoreXP"),
	[7] = Apollo.GetString("TargetFrame_IncreasedXP"),
	[8] = Apollo.GetString("TargetFrame_HighlyIncreasedXP"),
	[9] = Apollo.GetString("TargetFrame_GreatlyIncreasedXP"),
}

local _tRankDescription = {
	[_eRank.Fodder]		= Apollo.GetString("TargetFrame_VeryWeak"),
	[_eRank.Minion]		= Apollo.GetString("TargetFrame_Weak"),
	[_eRank.Standard]	= Apollo.GetString("TargetFrame_EasyAppend"),
	[_eRank.Champion]	= Apollo.GetString("TargetFrame_AlmostEqual"),
	[_eRank.Superior]	= Apollo.GetString("TargetFrame_Strong"),
	[_eRank.Elite]		= Apollo.GetString("TargetFrame_VeryStrong"),
}
--local _tRankDescriptionMeta = {
--	[_eRank.Fodder]		= Apollo.GetString("TargetFrame_Fodder"),
--	[_eRank.Minion]		= Apollo.GetString("TargetFrame_Minion"),
--	[_eRank.Standard]	= Apollo.GetString("TargetFrame_Grunt"),
--	[_eRank.Champion]	= Apollo.GetString("TargetFrame_Challenger"),
--	[_eRank.Superior]	= Apollo.GetString("TargetFrame_Superior"),
--	[_eRank.Elite]		= Apollo.GetString("TargetFrame_Prime"),
--}

local _tClassSprite = {
	[_eClass.Warrior]		= "IconSprites:Icon_Windows_UI_CRB_Warrior",
	[_eClass.Engineer]		= "IconSprites:Icon_Windows_UI_CRB_Engineer",
	[_eClass.Esper] 		= "IconSprites:Icon_Windows_UI_CRB_Esper",
	[_eClass.Medic] 		= "IconSprites:Icon_Windows_UI_CRB_Medic",
	[_eClass.Stalker] 		= "IconSprites:Icon_Windows_UI_CRB_Stalker",
	[_eClass.Spellslinger] 	= "IconSprites:Icon_Windows_UI_CRB_Spellslinger",
}

local _nTargetMax 			= 200			-- max distance to show if targeted. same as client max range for showing targeted models

local _nPoolLimit			= 100			-- the pool max size
local _nPoolMinLimit		= 15			-- the minimum window pool size
local _nReclamationAge		= 25			-- the number visi test ticks before window is reclaimed

local _L = {								-- rolled up, too many
	"Vulnerable",
	"Innate_Health",
	"Character_ShieldLabel",
	"FloatText_AbsorbTester",
	"TargetFrame_HealthFormat",
	"TargetFrame_HealthText",
	"TargetFrame_HealthShieldText",
	"CRB_Percent",
	"TargetFrame_IsPC",
	"TargetFrame_CreatureRank",
	"BaseBar_PvPAppend",
	"Adaptive",
	"TargetFrame_CreatureScales",
	"TargetFrame_TargetXPReward",
	"TargetFrame_Man",
	"TargetFrame_GroupSize",
	"TargetFrame_PvPMarked",
	"TargetFrame_PvPBonus",
	"TargetFrame_PvPCooldown",
	"TargetFrame_PvPOff",
	"TargetFrame_PvPForced",
	"TargetFrame_HealingAbsorbFormat",
	"Character_HealingAbsorbLabel",
}
for k,v in ipairs(_L) do
	_L[v] = Apollo.GetString(v); _L[k] = nil
end
_L["TargetView Options"] = "TargetFrame " .. Apollo.GetString("CRB_Options")

local _tPixie = {
	strSprite = "", cr = "",
	strText = nil, flagsText = {},
	loc = { fPoints = { 0, 0, 0, 0 }, nOffsets = { 0, 0, 0, 0 } }
}

local _floor = math.floor
local _sqrt = math.sqrt
local _abs = math.abs
local _format = string.format
local _len = string.len
local _tsort = table.sort
local _byte = string.byte
local _char = string.char

local function _dummy() end
local print = _dummy

local function _print(...)
	local s = ""
	for i = 1, select("#",...) do
		s = s .. tostring(select(i,...)) .. " "
	end
	Print(s)
end

local function DeepCopy(o)
	if type(o) ~= 'table' then return o end

	local copy = {}
	for k,v in next, o do copy[k] = DeepCopy(v) end
	return copy
end

local function SettingsCopy(o)
	local copy = {}
	for k,v in next, o do
		if type(v) == "table" and v.nControlType then v = v.default end
		copy[k] = DeepCopy(v)
	end
	return copy
end

local function PostChat(s)
	ChatSystemLib.PostOnChannel(ChatSystemLib.ChatChannel_System, s)
end

local _tShortFormatWhole = {
	[1] = Apollo.GetString("TargetFrame_ShortNumberWhole"),		-- thousands
	[2] = Apollo.GetString("TargetFrame_MillionsNumberWhole"),
	[3] = Apollo.GetString("TargetFrame_BillionsNumberWhole")
}

local _tShortFormatFloat = {
	[1] = Apollo.GetString("TargetFrame_ShortNumberFloat"),		-- thousands
	[2] = Apollo.GetString("TargetFrame_MillionsNumberFloat"),
	[3] = Apollo.GetString("TargetFrame_BillionsNumberFloat")
}

local function ShortNumber( n )							-- replaces FormatNumber/BigNumber, short number packed with thousands
	local i = 0											--   place indicator, for fast comparisons between frames.
	while n >= 1000 and i < 3 do						--   format: 54321 as int, where 1 is thousands root, and 543.2 is short form
		n = n / 1000									-- use with matching FormatNumber
		i = i + 1
	end
	return _floor((n - n % .1) * 100 + i)				-- floor to clip float errors
end

local function ShortRound( n )							-- replaces FormatNumber/BigNumber, short number packed with thousands rounded
	local i = 0											--   place indicator, for fast comparisons between frames.
	while n >= 1000 and i < 3 do						--   format: 54321 as int, where 1 is thousands root, and 543.2 is short form
		n = n / 1000									-- use with matching FormatNumber
		i = i + 1
	end
	n = n + 0.05; return _floor((n - n % .1) * 100 + i)	-- floor to clip float errors
end

local function FormatNumber( n )
	local i = n % 10									-- thousands root
	n = (n - i) / 100									-- short number form
	return i == 0 and tostring(n)
		or String_GetWeaselString((n % 1 == 0 and _tShortFormatWhole or _tShortFormatFloat)[i], n)
end

local function ShortTime( n )							-- replaces ConvertSecondsToTimer,
	local s = _floor(n)									--   for fast comparisons between frames
	return s											-- use with matching FormatTime
end

local function FormatTime( n )
	local s = n % 60
	local m = _floor(n % 3600 / 60)
	local h = _floor(n / 3600)
	return (h == 0  and "" or h .. ":") .. m .. ":" .. (s < 10 and (0 .. s) or s)
end

local function BuildTooltip(sTitle, sBody, csTitle)		-- _csTooltipBody = "ffc0c0c0", _csTooltipTitle = "ffdadada"
	if sTitle then
		if csTitle then
			return '<P Font="CRB_InterfaceMedium_B" TextColor="' .. csTitle .. '">' .. sTitle ..
				'</P><P><T Font="CRB_InterfaceMedium" TextColor="ffc0c0c0">' .. sBody .. '</T></P>'
		else
			return '<P Font="CRB_InterfaceMedium_B" TextColor="ffdadada">' .. sTitle ..
				'</P><P><T Font="CRB_InterfaceMedium" TextColor="ffc0c0c0">' .. sBody .. '</T></P>'
		end
	end
	return '<T Font="CRB_InterfaceMedium" TextColor="ffc0c0c0">' .. sBody .. '</T>'
end


-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------

function TargetView:new(o)
	o = self
	--o = o or {}
	--setmetatable(o, self)
	--self.__index = self

	o.L = o.L or _L or {}
	setmetatable(o.L, { __index = function(t,k) return tostring(k) end })

	o._AddonName = _AddonName
	o._defaultSettings = _defaultSettings
	o._resetSettings = _resetSettings
	o._print = _print

	o.arShowPlates = {}
	o.arNameplates = {}
	o.arUnit2Plate = {}

	o.arPreloadUnits = {}
	o.arPlatePool = {}
	o.arWindowPool = {}

	--o.unitPlayer = nil
	--o.unitTarget = nil
	--o.unitFocus = nil
	--o.bRequestRewards = nil
	--o.bPlayerFlagged = nil

	return o
end

function TargetView:Init()
	local bHasConfigureFunction = true
	local strConfigureButtonText = Apollo.GetString("OptionsHUD_TargetFrameLabel")
	local tDependencies = { "RewardIcons" }		-- { "Tooltips", "RewardIcons" }
	Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end

function TargetView:OnDependencyError(strDependency, strError)
	return true
end


-----------------------------------------------------------------------------------------------
-- TargetView OnLoad
-----------------------------------------------------------------------------------------------

function TargetView:OnLoad()
	self.onXmlDocLoadedCalled = false

	self.userSettings = SettingsCopy(_defaultSettings)

	-- create form
	self.xmlDoc = XmlDoc.CreateFromFile("TargetView.xml")
	self.xmlDoc:RegisterCallback("OnDocLoaded", self)

	Apollo.RegisterEventHandler("UnitCreated", "OnUnitEnter", self)	-- unit events can get created before docload

	Apollo.RegisterEventHandler("WindowManagementReady", "OnWindowManagementReady", self)
	Apollo.RegisterEventHandler("TargetView_ShowMenu", "OnShowMenu", self)
	Apollo.RegisterEventHandler("InterfaceMenuListHasLoaded", "OnInterfaceMenuListHasLoaded", self)
	Apollo.RegisterSlashCommand("TargetFrame", "OnCommand", self)
	Apollo.RegisterSlashCommand("targetframe", "OnCommand", self)
	Apollo.RegisterSlashCommand("TargetView", "OnCommand", self)
	Apollo.RegisterSlashCommand("targetview", "OnCommand", self)
	Apollo.RegisterSlashCommand("focus", "OnFocusSlashCommand", self)
end

function TargetView:OnWindowManagementReady()
	Event_FireGenericEvent("WindowManagementRegister", {strName=self.L["TargetView Options"], nSaveVersion=_VersionWindow})
	for _,v in next, _ePlateType do
		Event_FireGenericEvent("WindowManagementRegister", {strName=_tWindowManagementLabel[v], nSaveVersion=2})
	end
end

function TargetView:OnInterfaceMenuListHasLoaded()
	Event_FireGenericEvent("InterfaceMenuList_NewAddOn", _AddonName,
		{ "TargetView_ShowMenu", "", "zonemap:UI_WorldMap_Position" })
end


-----------------------------------------------------------------------------------------------
-- TargetView OnDocLoaded
-----------------------------------------------------------------------------------------------

function TargetView:OnDocLoaded()
	if self.xmlDoc == nil or not self.xmlDoc:IsLoaded() then
		return
	end

	Event_FireGenericEvent("OneVersion_ReportAddonInfo", _AddonName, 1, _Version, _VersionMinor)

	-- Load the window

	-- Apollo.RegisterEventHandler("UnitCreated", 					"OnUnitEnter", self)
	Apollo.RegisterEventHandler("UnitDestroyed",				"OnUnitEnter", self)
	Apollo.RegisterEventHandler("NextFrame",					"OnFrame", self)

	Apollo.RegisterEventHandler("CharacterCreated",				"OnCharacterLoaded", self)
	Apollo.RegisterEventHandler("TargetUnitChanged",			"OnTargetUnitChanged", self)
	Apollo.RegisterEventHandler("AlternateTargetUnitChanged",	"OnAlternateTargetUnitChanged", self)
	Apollo.RegisterEventHandler("OptionsUpdated_UnitFramesUpdated",	"OnSettingRebuildFrames", self)
	Apollo.RegisterEventHandler("Tutorial_RequestUIAnchor",		"OnTutorial_RequestUIAnchor", self)
	Apollo.RegisterEventHandler("KeyBindingKeyChanged", 		"OnKeyBindingUpdated", self)
	Apollo.RegisterEventHandler("ResolutionChanged",			"OnResolutionChanged", self)

	local t = {
		"QuestObjectiveUpdated", "QuestStateChanged", "ChallengeAbandon", "ChallengeLeftArea", "ChallengeFailTime",
		"ChallengeFailArea", "ChallengeActivate", "ChallengeCompleted", "ChallengeFailGeneric", "PublicEventObjectiveUpdate",
		"PublicEventUnitUpdate", "PlayerPathMissionUpdate", "FriendshipAdd", "FriendshipPostRemove", "FriendshipUpdate",
		"PlayerPathRefresh", "ContractObjectiveUpdated", "ContractStateChanged", "ChallengeUpdated",
	}
	for _,v in pairs(t) do
		Apollo.RegisterEventHandler(v, "OnRequestRewardsUpdate", self)
	end

	self.tUpdateTimer = ApolloTimer.Create(0.5, true, "OnUpdateTimer", self)

	self.onXmlDocLoadedCalled = true
	self:OnUnitFrameOptionsUpdated()			--FIXME: generates all three windows. need to finish dynamic allocation, this removed here
	self:InitializeWindows()
end


-----------------------------------------------------------------------------------------------
-- TargetView Functions
-----------------------------------------------------------------------------------------------

function TargetView:OnFrame()
	--local nTime = GameLib.GetTickCount()													--PROF:
	--self._frameCount = (self._frameCount or 0) + 1										--PROF:
	--local ou = self.userSettings

	for _,plate in next, self.arShowPlates do

		local unit = plate.unitOwner

		self:DrawHealth(plate)
		self:DrawInterruptShield(plate)
		self:UpdateCastingBar(plate)
		--FIXME: maybe put actual ToT target switch here in fast checks

		-- marker
		local nMarkerId = unit:GetTargetMarker() or 0
		if nMarkerId ~= plate.nMarkerId then
			plate.marker:SetSprite(nMarkerId == 0 and "" or _tMarkerToSprite[nMarkerId])
			plate.marker:Show(nMarkerId ~= 0, true)
			plate.nMarkerId = nMarkerId
		end

		-- cluster plates
		for _,cluster in next, plate.arShowClusters do
			self:DrawHealthCluster(cluster)
			self:DrawCastCluster(cluster)
		end
	end
	--if self._frameCount % 30 == 1 then													--PROF:
	--	Print((self._frameCount or 0) .. ": " .. (GameLib.GetTickCount() - nTime))			--PROF:
	--end																					--PROF:
end

function TargetView:DrawHealth(plate)
	local o,ou,wf,unit = plate.userSettings, self.userSettings, plate.frame, plate.unitOrVehicle

	local nHealthMax = unit and unit:GetMaxHealth() or 0
	local nHealth    = nHealthMax == 0 and 0 or unit:GetHealth()

	-- color
	local ratioH = nHealthMax == 0 and 1 or nHealth / nHealthMax
	local bShow = (ratioH <= _nHealthRed and _eTintType.Critical)
		or (ratioH <= _nHealthYellow and _eTintType.Warning)
		or _eTintType.Good

	if bShow ~= plate.eHealthTintType then
		plate.healthFill:SetFullSprite(plate._tHealthBarSprite[bShow])
		plate.eHealthTintType = bShow
	end

	bShow = nHealth > 0 and ratioH < _nHealthGlow
	if bShow ~= plate.bShowHealthGlow then
		plate.healthFill:SetStyleEx("EdgeGlow", bShow)
		plate.bShowHealthGlow = bShow
	end

	-- simple
	bShow = nHealthMax == 0 or plate.sUnitType == "Simple"
	if bShow ~= plate.bIsSimple then
		if bShow then
			plate.shield:Show(false)
			plate.bShowShield = false
			plate.absorb:Show(false)
			plate.bShowAbsorb = false
			plate.healthText:SetText("")
			plate.sBarText = ""
			plate.healthText:SetTooltip("")
			plate.sTooltipText = ""
			plate.nHealthMax = nil
		end
		plate.bIsSimple = bShow
	end
	if bShow then
--self._cSkip = (self._cSkip or 0) + 1													--PROF:
--Print("skip " .. tostring(nHealthMax) .. " " .. tostring(unit:GetName()))				--PROF:
		return
	end

	local nShieldMax  = unit:GetShieldCapacityMax() or 0
	local nShield     = nShieldMax == 0 and 0 or unit:GetShieldCapacity()
	local nAbsorbMax  = unit:GetAbsorptionMax() or 0
	local nAbsorb     = nAbsorbMax == 0 and 0 or unit:GetAbsorptionValue()	-- doesn't clear itself when absorb buff drops off
	local nHealAbsorb = unit:GetHealingAbsorptionValue() or 0
	local nHealMin    = unit:GetHealthFloor() or 0
	local nHealMax    = unit:GetHealthCeiling() or 0
	--local nHealAbsorb = (unit:GetHealingAbsorptionValue() or 0) + nHealthMax - nHealth
	--local nHealMin    = (unit:GetHealthFloor() or 0) + 1
	--local nHealMax    = (unit:GetHealthCeiling() or 0) + nHealthMax - 1

	local bNewHealth = nHealth ~= plate.nHealth or nHealthMax ~= plate.nHealthMax
	local bNewShield = nShield ~= plate.nShield or nShieldMax ~= plate.nShieldMax
	local bNewAbsorb = nAbsorb ~= plate.nAbsorb or nAbsorbMax ~= plate.nAbsorbMax
	local bNewHeal = nHealAbsorb ~= plate.nHealAbsorb or nHealMin ~= plate.nHealMin or nHealMax ~= plate.nHealMax
	local eTextDisplay = ou.eTextDisplay
	local bNewOption = eTextDisplay ~= plate.eTextDisplay

	if not bNewHealth and not bNewShield and not bNewAbsorb and not bNewHeal and not bNewOption then
		return
	end

	-- bar window position
	local bShowShield = nShield > 0
	local bShowAbsorb = nAbsorb > 0

	if bShowShield ~= plate.bShowShield or bShowAbsorb ~= plate.bShowAbsorb then
--Print("pos: " .. tostring(plate.plateType))											--PROF:
		if bShowShield ~= plate.bShowShield then
			plate.shield:Show(bShowShield)
			plate.absorb:MoveToLocation(bShowShield and plate.arAbsorbPos or plate.arShieldPos)
			plate.bShowShield = bShowShield
		end

		if bShowAbsorb ~= plate.bShowAbsorb then
			plate.absorb:Show(bShowAbsorb)
			plate.bShowAbsorb = bShowAbsorb
		end

		local bBoth,bOne = bShowShield and bShowAbsorb, bShowShield or bShowAbsorb
		if bOne ~= plate.bShowHealthSplit then
			plate.healthSplit:Show(bOne)
			plate.bShowHealthSplit = bOne
		end

		local ofs = (bBoth and _nFrameOfsBoth) or (bOne and _nFrameOfsOne) or _nFrameOfsNone	-- resize bars
		local p,c,x =  plate._nFramePoints, plate.tCastClipPoints, plate._tCastClipPointsX
		if o.bFlipped then
			wf:SetAnchorOffsets(p[1] + ofs, p[2], p[3], p[4])
			c[1].nX, c[2].nX, c[3].nX, c[4].nX  = x[1] + ofs, x[2] + ofs, x[3], x[4]	-- TL, BL, BR, TR
		else
			wf:SetAnchorOffsets(p[1], p[2], p[3] - ofs, p[4])
			c[1].nX, c[2].nX, c[3].nX, c[4].nX  = x[1], x[2], x[3] - ofs, x[4] - ofs	-- TL, BL, BR, TR
			plate.bRequestLayout = true				-- right facing plate with icons toward outside, re-layout
		end
		plate.castFill:SetClipPoints(c[1], c[2], c[3], c[4])
	end

	local nHealthMaxAbsorb = nHealthMax + nHealAbsorb

	-- heal absorb
	bShow = nHealAbsorb > 0
	if bShow then
		if nHealthMaxAbsorb ~= plate.nHealthMaxAbsorb then
			plate.healAbsorb:SetMax(nHealthMaxAbsorb)
			plate.nHealthMaxAbsorb = nHealthMaxAbsorb
		end
		local nHealAbsorbMark = nHealth + nHealAbsorb
		if nHealAbsorbMark ~= plate.nHealAbsorbMark then
			plate.healAbsorb:SetProgress(nHealAbsorbMark)
			plate.nHealAbsorbMark = nHealAbsorbMark
		end
	end
	if bShow ~= plate.bShowHealAbsorb then
		plate.healAbsorb:Show(bShow)
		plate.bShowHealAbsorb = bShow
	end
	
	-- heal clamp
	bShow = nHealMin > 0
	if bShow and (nHealthMaxAbsorb ~= plate.nHealMinMax or nHealMin ~= plate.nHealMin) then
		plate.healMin:SetMax(nHealthMaxAbsorb)						-- set max before progress or control breaks
		plate.healMin:SetProgress(nHealMin)
		plate.nHealMinMax = nHealthMaxAbsorb
	end
	if bShow ~= plate.bShowHealMin then
		plate.healMin:Show(bShow)
		plate.bShowHealMin = bShow
	end
	bShow = nHealMax > 0 and nHealMax ~= nHealthMax 
	if bShow and (nHealthMaxAbsorb ~= plate.nHealMaxMax or nHealMax ~= plate.nHealMax) then
		plate.healMax:SetMax(nHealthMaxAbsorb)						-- set max before progress or control breaks
		plate.healMax:SetProgress(nHealMax)
		plate.nHealMaxMax = nHealthMaxAbsorb
	end
	if bShow ~= plate.bShowHealMax then
		plate.healMax:Show(bShow)
		plate.bShowHealMax = bShow
	end

	if bNewHeal then
		plate.nHealAbsorb, plate.nHealMin, plate.nHealMax = nHealAbsorb, nHealMin, nHealMax
	end

	-- bar progress
	if nHealthMaxAbsorb ~= plate.nHealthMaxMax then
		plate.healthFill:SetMax(nHealthMaxAbsorb)
		plate.nHealthMaxMax = nHealthMaxAbsorb
	end
	if nHealth ~= plate.nHealth then
		plate.healthFill:SetProgress(nHealth)
		plate.nHealth = nHealth
	end

	-- shield
	if nShieldMax ~= plate.nShieldMax then		-- shield and absorb are progress fills
		plate.shieldFill:SetMax(nShieldMax)		-- set max before progress or control breaks
		plate.nShieldMax = nShieldMax
	end
	if nShield ~= plate.nShield then
		plate.shieldFill:SetProgress(nShield)
		plate.nShield = nShield
	end

	-- absorb
	if nAbsorbMax ~= plate.nAbsorbMax then
		plate.absorbFill:SetMax(nAbsorbMax)
		plate.nAbsorbMax = nAbsorbMax
	end
	if nAbsorb ~= plate.nAbsorb then
		plate.absorbFill:SetProgress(nAbsorb)
		plate.nAbsorb = nAbsorb
	end

	-- tooltip text
	local s,f = nil, String_GetWeaselString

	if bNewHealth then
		local n,m,r = ShortNumber(nHealth), ShortNumber(nHealthMax), _floor(ratioH * 100)
		if n ~= plate.nHealthClip or m ~= plate.nHealthMaxClip or r ~= plate.nHealthRatio then
			plate.sHealthTooltip = f(_L.TargetFrame_HealthFormat, _L.Innate_Health,
				FormatNumber(n), FormatNumber(m), f(_L.CRB_Percent, r))
			plate.nHealthClip, plate.nHealthMaxClip, plate.nHealthRatio = n,m,r
		else
			bNewHealth = false			-- for the bar text test below
		end
		plate.nHealthMax = nHealthMax
	end
	if bNewShield then
		local n,m,r = ShortNumber(nShield), ShortNumber(nShieldMax), _floor(nShield / nShieldMax * 100)
		if n ~= plate.nShieldClip or m ~= plate.nShieldMaxClip or r ~= plate.nShieldRatio then
			plate.sShieldTooltip = f(_L.TargetFrame_HealthFormat, _L.Character_ShieldLabel,
				FormatNumber(n), FormatNumber(m), f(_L.CRB_Percent, r))
			plate.nShieldClip, plate.nShieldMaxClip, plate.nShieldRatio = n,m,r
		else
			bNewShield = false
		end
	end
	if bNewAbsorb then
		local n,m,r = ShortNumber(nAbsorb), ShortNumber(nAbsorbMax), _floor(nAbsorb / nAbsorbMax * 100)
		if n ~= plate.nAbsorbClip or m ~= plate.nAbsorbMaxClip or r ~= plate.nAbsorbRatio then
			plate.sAbsorbTooltip = f(_L.TargetFrame_HealthFormat, _L.FloatText_AbsorbTester,
				FormatNumber(n), FormatNumber(m), f(_L.CRB_Percent, r))
			plate.nAbsorbClip, plate.nAbsorbClip, plate.nAbsorbRatio = n,m,r
		end
	end
	if bNewHeal then
		local n = ShortNumber(nHealAbsorb)
		if n ~= plate.nHealAbsorbClip then
			plate.sHealAbsorbTooltip = f(_L.TargetFrame_HealingAbsorbFormat, _L.Character_HealingAbsorbLabel,
				FormatNumber(n))
			plate.nHealAbsorbClip = n
		end
	end

	s = plate.sHealthTooltip						-- 100% always
	if bShowShield then								-- 100% non-combat, 95%+ combat
		s = s .. "\n" .. plate.sShieldTooltip
	end
	if bShowAbsorb then								-- 99% rare-ish, dont try to complicate the shield concat
		s = s .. "\n" .. plate.sAbsorbTooltip
	end
	if plate.bShowHealAbsorb then
		s = s .. "\n" .. plate.sHealAbsorbTooltip
	end
	if s ~= plate.sTooltipText then					-- happens about 90%+ less often than the string builds,
		plate.healthText:SetTooltip(s)				--   as the trunced values result in same string
		plate.sTooltipText = s
	end

	-- bar text
	s = ""											-- on mouseover, default setting so gets fast path
	if eTextDisplay == 2 then						-- show x / y
		if bNewHealth or bNewShield or bNewOption then
			s = f(_L.TargetFrame_HealthText, FormatNumber(plate.nHealthClip), FormatNumber(plate.nHealthMaxClip))
			if bShowShield then
				s = f(_L.TargetFrame_HealthShieldText, s, FormatNumber(plate.nShieldClip))
			end
		end
	elseif eTextDisplay == 3 then					-- show %
		if bNewHealth or bNewShield or bNewOption then
			s = f(_L.CRB_Percent, plate.nHealthRatio)
			if bShowShield and plate.nHealthRatio ~= plate.nShieldRatio then
				s = f(_L.TargetFrame_HealthShieldText, s, f(_L.CRB_Percent, plate.nShieldRatio))
			end
		end
	end
	if s ~= plate.sBarText then
		plate.healthText:SetText(s)
		plate.sBarText = s
	end
	if bNewOption then
		plate.eTextDisplay = eTextDisplay
	end
end

function TargetView:DrawInterruptShield(plate)		--FIXME: consider running this function first cause GetInterruptArmorMax()
	local wf,unit,player,unitOrVehicle = plate.frame, plate.unitOwner, self.unitPlayer, plate.unitOrVehicle

	local eIaState, nCCArmorValue = (_eCastState.Unknown)

	local nVulnerableTime = plate.unitOrVehicle:GetCCStateTimeRemaining(_eStateVulnerability) or 0
	if nVulnerableTime ~= plate.nVulnerableTime then
		plate.nVulnerableTime = nVulnerableTime
	end

	if nVulnerableTime > 0 then				-- vuln overrides all and must be checked
		eIaState = _eCastState.MoO
	else
		local nCCArmorMax = unit:GetInterruptArmorMax()

		if nCCArmorMax == 0 then					-- 99%, one empty check
		elseif nCCArmorMax == -1 then
			eIaState = _eCastState.Locked
		else										--if nCCArmorMax > 0 then
			nCCArmorValue = unit:GetInterruptArmorValue()
			if nCCArmorValue > 0 then
				eIaState = _eCastState.Value
			end
		end
	end

	local bShow = eIaState ~= _eCastState.Unknown

	if eIaState ~= plate.eIaState then
		plate.ccArmor:SetSprite(_tInteruptArmorSprite[eIaState])
		plate.eIaState = eIaState
	end
	if bShow and nCCArmorValue ~= plate.nCCArmorValue then
		plate.ccArmor:SetText(nCCArmorValue)
		plate.nCCArmorValue = nCCArmorValue
	end

	if bShow ~= plate.bShowCCArmor then
		plate.ccArmor:Show(bShow)
		plate.bShowCCArmor = bShow
	end
end

function TargetView:UpdateCastingBar(plate)
	local unit = plate.unitOrVehicle

	local nVulnerableTime, eCastType = (plate.nVulnerableTime)

	local bShow = nVulnerableTime > 0
	if not bShow and unit:ShouldShowCastBar() then
		eCastType = unit:GetCastBarType()
		bShow = eCastType and eCastType ~= _eCastType.None
	end

	if bShow ~= plate.bShowCasting then
		plate.castBar:Show(bShow)
		plate.bShowCasting = bShow
	end

	if not bShow then
		return
	end

	local o,wb,wn = plate.userSettings, plate.castFill, plate.castName

	if eCastType ~= plate.eCastType then
		plate.castIcon:SetSprite(_tCastIconSprite[eCastType])
		plate.eCastType = eCastType
	end

	local bIsVuln = nVulnerableTime > 0
	if bIsVuln ~= plate.bIsVuln then
		wb:EnableGlow(not bIsVuln)
		plate.bIsVuln = bIsVuln
	end

	local crCastText = "magenta"			--FIXME: could still clean this section up more
	local eCastState = _eCastState.MoO
	if not bIsVuln then
		local bIsLocked = unit:GetInterruptArmorMax() == -1
		crCastText = bIsLocked and "CoolGrey" or "PastelOrange"
		eCastState = bIsLocked and _eCastState.Locked or _eCastState.Value
	end
	if crCastText ~= plate.crCastText then
		wn:SetTextColor(crCastText)
		plate.crCastText = crCastText
	end
	if eCastState ~= plate.eCastState then
		local bFlipped = o.bFlipped
		wb:SetFullSprite((bFlipped and _tCastBarFillSpriteFlipped or _tCastBarFillSprite)[eCastState])
		plate.castBase:SetSprite((bFlipped and _tCastBarSpriteFlipped or _tCastBarSprite)[eCastState])
		plate.eCastState = eCastState
	end

	local nDuration = bIsVuln and 4000 or unit:GetCastDuration()
	if nDuration ~= plate.nDuration then
		wb:SetMax(nDuration)
		plate.nDuration = nDuration
	end
	local nElapsed = bIsVuln and nVulnerableTime * 1000 or unit:GetCastElapsed()
	if nElapsed ~= plate.nElapsed then
		wb:SetProgress(nElapsed)
		plate.nElapsed = nElapsed
	end

	local sSpellName = bIsVuln and _L.Vulnerable or unit:GetCastName()
	local nCastRemain = bIsVuln and nVulnerableTime or (nDuration - nElapsed) / 1000
	nCastRemain = nCastRemain - (nCastRemain % .1)
	if nCastRemain ~= plate.nCastRemain or sSpellName ~= plate.sSpellName then
		local s = sSpellName
		if nDuration > 999 then				-- add a countdown timer if nDuration is > 0.999 seconds
			local n = nCastRemain % 1
			s = s .. " (" .. nCastRemain - n .. "." .. n * 10 .. ")"
		end
		wn:SetText(s)
		plate.sSpellName = sSpellName
		plate.nCastRemain = nCastRemain
	end
end


-- create destroy		-----------------------------------------------------------------------

function TargetView:OnUnitEnter(unit)		-- Created/Detroyed, temp name for file comparison
	local player = self.unitPlayer

	if not player or not player:IsValid() then
		player = GameLib.GetPlayerUnit()
		self.unitPlayer = player
	end

	local target = self.unitTarget
	if not target or not target:IsValid() then
		target = GameLib.GetTargetUnit()
		self.unitTarget = target
	end

	local focus = self.unitFocus
	if player and (not focus or not focus:IsValid()) then
		focus = player:GetAlternateTarget()
		self.unitFocus = focus
	end
end

function TargetView:OnUnitDestroyed(plate)
--self._profE = (self._profE or 0) + 1											--PROF:
	local s2p,i2p,u2p = self.arShowPlates, self.arNameplates, self.arUnit2Plate

	local plateType = plate.plateType
	if not plate then
		return
	end

	local w,pool = plate.nameplate, self.arPlatePool
	local icount,scount,limit = #i2p, #s2p, _nPoolLimit

	plate.unit = nil
	plate.unitOrVehicle = nil

	if w then							-- check window
		plate.bShow = false
		w:Show(false, true)
		w:SetData(nil)
		if plate.rewards then					--FIXME: rewards isnt set sometimes even though w is, mystery
			plate.rewards:SetData(nil)			-- release RewardIcons table tied to unit
		end

		--limit = limit - (icount - scount)		-- constrain by inverse of plates to visible
		--if limit < _nPoolMinLimit then limit = _nPoolMinLimit end
		--pool = self.arWindowPool
	end

	local count					-- check pool
	if count then
		--pool[count + 1] = plate
	elseif w then
		w:Destroy()
--self._profWd = (self._profWd or 0) + 1											--PROF:
--self._profPd = (self._profPd or 0) + 1											--PROF:
--	else																			--PROF:
--self._profPd = (self._profPd or 0) + 1											--PROF:
	end

	u2p[plateType] = nil					-- remove from indexes

	for i = 1,icount do
		if i2p[i] == plate then
			i2p[i] = i2p[icount]; i2p[icount] = nil
			break
		end
	end

	for i = 1,scount do
		if s2p[i] == plate then
			s2p[i] = s2p[scount]; s2p[scount] = nil
--self._profCr = (self._profCr or 0) + 1										--PROF:
			break
		end
	end
end

function TargetView:OnUnitCreated(plateType)
--self._profE = (self._profE or 0) + 1											--PROF:
--local unitType = unit:GetType()
	--Print("create: " .. tostring(unit:GetName()))								--PROF:
	--	.. " age:" .. plate.nAge .. " now:" .. self.nAge)						--PROF:

	local i2p,u2p,pool = self.arNameplates, self.arUnit2Plate, self.arPlatePool

	local plate = u2p[plateType]
	if plate then
--Print("has")																	--PROF:
		return
	end

	--local count = #pool
	--if count > 0 then												-- check plate pool
	--	plate, pool[count] = pool[count], nil
	--else
		plate = { }
--self._profPc = (self._profPc or 0) + 1										--PROF:
	--end

	-- set plate

	self:InitPlate(plate, plateType, unit, unitType)

	i2p[#i2p + 1] = plate
	u2p[plateType] = plate

	self:PromotePlate(plate)			--FIXME: temporary, until cluster show is split up correctly

	plate.bRequestInit = true			-- may never happen
	plate.bRequestRewards = true		-- needs just one independant update if testing gets far enough
end

function TargetView:InitPlate(plate, plateType, unit, unitType)
	local ou = self.userSettings

	if not self.tPlateTemplate then		-- make static template with base values
		self.tPlateTemplate = { }
		self:InitPlate(self.tPlateTemplate, plateType, unit, unitType)

--Print("----")																		--PROF:
--for k,v in next, self.tPlateTemplate do											--PROF:
--	Print(" " .. k .. ": " .. tostring(v))											--PROF:
--end																				--PROF:
	end

	--local guild = self.tPlayerGuild
	--local guild, warparty = self.tPlayerGuild, self.tPlayerWarparty

	plate.plateType 			= plateType
	plate.unitOwner				= unit
	plate.sUnitType				= unitType

	local o = { }										-- options
	plate.userSettings = o
	o.fScale 		= 1.0
	o.cvkVisibility	= nil
	o.bShowClusters = true
	o.bShowToT		= true
	o.bFlipped		= false
	o.bFlipClusters = false

	local ePlateType = _ePlateType[plateType]

	local bFlipped = (g_InterfaceOptions and g_InterfaceOptions.Carbine or {})[_tCarbineOptionFlip[ePlateType]]
	if bFlipped then
		o.bFlipped = bFlipped
	end

	if plateType == "unitPlayer" then
		o.bShowToT = false
		o.cvkVisibility = "hud.myUnitFrameDisplay"
		o.bShowClusters = ou.bShowClusters and ou.bShowMyCluster
	elseif plateType == "unitTarget" then
		o.bFlipClusters = true
		o.bShowClusters = ou.bShowClusters and ou.bShowTargetCluster
	elseif plateType == "unitFocus" then
		o.cvkVisibility = "hud.focusTargetFrameDisplay"
		o.bShowClusters = ou.bShowClusters and ou.bShowFocusCluster
	end

	plate.locDefaultPosition = _tDefaultPosition[ePlateType]

	plate.nLastCCArmorValue		= 0
	plate.bTargetDead			= false

		-- re-set in event checks
	--plate.unitOwner	= unit

	-- dependant on window state:   so leave as is or default nil
		-- re-set in slow 'visibility' checks
	plate.nAge					= 0
	--plate.unitOrVehicle		= unit
	--plate.bShow				= nil
	--plate.bShowClassIcon		= nil
	--plate.eClassId			= nil
	--plate.bShowCertainDeath	= nil
	--plate.bIsRare				= nil
	--plate.bIsElite			= nil

		-- re-set in fast checks
	--plate.bShowToT			= nil
	--plate.nMarkerId			= nil
	--plate.bShowCasting		= nil
	--plate.eCastType			= nil
	--plate.nElapsed			= nil
	--plate.nDuration			= nil
	--plate.crCastText			= nil
	--plate.eDispositionToT		= nil
	--plate.nVulnerableTime		= nil								-- total time
	--plate.nCCArmorMax			= nil
	--plate.nCCArmorValue		= nil
	--plate.bShowCCArmor		= nil
	--plate.nHealth				= nil
	--plate.nHealthMax			= nil
	--plate.nShield				= nil
	--plate.nShieldMax			= nil
	--plate.nAbsorb				= nil
	--plate.nAbsorbMax			= nil

	-- clusters
	plate.arShowClusters = { }
	plate.arClusterplates = { }
	for i = 1,4 do
		self:OnClusterCreated(plate)
	end

	plate._tHealthBarSprite = o.bFlipped and _tHealthBarFlipped or _tHealthBar
	plate._tRareSprite = o.bFlipped and _tRareSpriteFlipped or _tRareSprite
	plate._tEliteSprite = o.bFlipped and _tEliteSpriteFlipped or _tEliteSprite
	plate._tSplitSprite = o.bFlipped and _tSplitSpriteFlipped or _tSplitSprite
	plate._tBackerSprite = o.bFlipped and _tBackerSpriteFlipped or _tBackerSprite
	plate._tAttachmentSprite = o.bFlipped and _tAttachmentSpriteFlipped or _tAttachmentSprite
end

function TargetView:DemotePlate(plate)
	local w,pool = plate.nameplate, self.arPlatePool

--	local count = #pool
--	local p = pool[count]
--	if not p then				-- no clean plates, leave it promoted
--		return
--	end
--Print("demote " .. (plate.plateType or "unknown") .. " " .. (plate.sUnitName or "unknown"))	--PROF:

	local wpool,ptemplate = self.arWindowPool, self.tPlateTemplate
	local wcount = #wpool

	w:SetData(nil)

	--plate.beneBuffBar:SetUnit(nil)
	--plate.harmBuffBar:SetUnit(nil)
	--plate.model:SetCostume(nil)

--	if wcount < _nPoolMinLimit then					-- save window
--		for k,v in next, plate do
--			if ptemplate[k] == nil then plate[k] = nil; p[k] = v end
--		end
--		pool[count], wpool[wcount + 1] = nil, p		-- change pool
--	else
		for _,v in next, plate.arClusterplates do	-- parent destroy, clear window dependent vars for clusters
			self:DemoteCluster(v)
		end
		for k,_ in next, plate do
			if ptemplate[k] == nil then plate[k] = nil end
		end
		w:Destroy()
--self._profWd = (self._profWd or 0) + 1											--PROF:
--	end

	Event_FireGenericEvent("WindowManagementRemove", {strName=_tWindowManagementLabel[_ePlateType[plate.plateType]]})

	plate.bRequestInit = true
end

function TargetView:PromotePlate(plate)
	local o,w,wf,wt,pool = plate.userSettings, plate.nameplate, plate.frame, plate.totFrame, self.arWindowPool
	--Print("promote " .. plate.unitOwner:GetName())								--PROF:
	if w then									-- curently not ok to call this if already promoted
		return									--   check added to let RequestInit flag re-init certain state
	end

	local count = #pool						-- check pool
	if count > 0 then
	--	local p = pool[count]				-- window pool
	--	pool[count] = nil

	--	pool = self.arPlatePool				-- plate pool
	--	count = #pool
	--	if count < _nPoolLimit then
	--		pool[count + 1] = p
----		else																		--PROF:
----self._profPd = (self._profPd or 0) + 1												--PROF:
	--	end

	--	for k,v in next, p do				-- transfer plate data
	--		if plate[k] == nil then plate[k] = v; p[k] = nil end
	--	end

	--	w = plate.nameplate
	end

	if not w or not w:IsValid() then
		w = Apollo.LoadForm(self.xmlDoc, o.bFlipped and "TargetFlipped" or "Target", "FixedHudStratumLow", self)
self._profWc = (self._profWc or 0) + 1

		plate.nameplate = w

		wf = w:FindChild("Frame")
		plate.frame = wf

		plate.healthSplit = wf:FindChild("HealthSplit")			-- health bar
		plate.health = wf:FindChild("Health")
		plate.shield = wf:FindChild("Shield")
		plate.absorb = wf:FindChild("Absorb")
		plate.healthFill = wf:FindChild("HealthFill")
		plate.shieldFill = wf:FindChild("ShieldFill")
		plate.absorbFill = wf:FindChild("AbsorbFill")
		plate.healthText = wf:FindChild("HealthText")
		plate.healAbsorb = wf:FindChild("HealAbsorbFill")
		plate.healMin = wf:FindChild("HealMin")
		plate.healMax = wf:FindChild("HealMax")
		plate.certainDeath = wf:FindChild("CertainDeath")
		plate.model = wf:FindChild("Model")
		plate.level = wf:FindChild("Level")
		plate.scalingMark = wf:FindChild("ScalingMark")
		plate.groupSizeIcon = wf:FindChild("GroupSizeIcon")
		plate.groupSizeMark = wf:FindChild("GroupSizeMark")
		plate.rewards = wf:FindChild("GoalPanel")
		plate.marker = wf:FindChild("Marker")
		plate.name = wf:FindChild("Name")
		plate.rare = wf:FindChild("Rare")
		plate.elite = wf:FindChild("Elite")
		plate.ccArmor = wf:FindChild("CCArmor")
		plate.beneBuffBar = wf:FindChild("BeneBuffBar")
		plate.harmBuffBar = wf:FindChild("HarmBuffBar")
		plate.classShadow = wf:FindChild("ClassShadow")
		plate.classIcon = wf:FindChild("ClassIcon")
		plate.attachment = wf:FindChild("Attachment")
		plate.backer = wf:FindChild("Backer")
		plate.clusters = wf:FindChild("Clusters")
		plate.pvpMarker = wf:FindChild("PvpMarker")
--Print("pvp: " .. tostring(plate.pvpMarker))										--PROF:

		plate.castBar = wf:FindChild("CastBar")			-- cast bar
		plate.castFill = wf:FindChild("CastFill")
		plate.castName = wf:FindChild("CastName")
		plate.castIcon = wf:FindChild("CastIcon")
		plate.castBase = wf:FindChild("CastBase")

		wt = w:FindChild("TotFrame")
		plate.totFrame = wt
		plate.totBacker = wt:FindChild("Backer")
		plate.totDisposition = wt:FindChild("DispositionFrame")
		plate.totModel = wt:FindChild("Model")
	end

	if not plate.pvpMarker then									--NOTE: work around for C side race condition
		local ptemplate = self.tPlateTemplate					--  by the Lua code, this state should never happen but does.
--		for k,_ in next, plate do								--  only happens for breadth of one frame for only one unit, and is excedingly rare
--			if ptemplate[k] == nil then plate[k] = nil end		--IMPORTANT: make sure this stays in sync with demote code
--		end
--		w:Destroy()
--		return													-- caller will attempt to promote again next frame
	end

	-- set window
	--plate.bShow = nil
	w:SetData(plate)
	--w:SetUnit(plate.unitOrMount, 1)

	-- one time fix-ups

--	plate.clusters:ArrangeChildrenHorz(o.bFlipClusters and
--		Window.CodeEnumArrangeOrigin.RightOrBottom or Window.CodeEnumArrangeOrigin.LeftOrTop)

	plate._nFramePoints = { wf:GetAnchorOffsets() }

	local cp = plate.castFill:GetClipPoints()
	plate.tCastClipPoints = cp
	plate._tCastClipPointsX = { cp[1].nX, cp[2].nX, cp[3].nX, cp[4].nX }

	plate.arShieldPos = plate.shield:GetLocation()
	plate.arAbsorbPos = plate.absorb:GetLocation()

	--self:UpdateLayout(plate)											-- so can load default pos
--[[
	w:SetScale(o.fScale)

	-- most of this needs to be removed

	--self:ArrangeClusterMembers(plate)
	w:ArrangeChildrenHorz(Window.CodeEnumArrangeOrigin.Middle)

	plate._nFramePoints = { wf:GetAnchorOffsets() }

	local cp = plate.castFill:GetClipPoints()
	plate.tCastClipPoints = cp
	plate._tCastClipPointsX = { cp[1].nX, cp[2].nX, cp[3].nX, cp[4].nX }

	plate.arShieldPos = plate.shield:GetLocation()
	plate.arAbsorbPos = plate.absorb:GetLocation()

	plate.shieldFill:SetMax(100)
	plate.shieldFill:SetFloor(0)
	plate.shieldFill:SetProgress(100)
]]--
	local locNewLocation = WindowLocation.new(plate.locDefaultPosition)	 -- so WindowManagementAdd starts with the right settings
	if locNewLocation then
		w:MoveToLocation(locNewLocation)
	end

	self:InitPromoted(plate)									--FIXME: remove on completing demote

	Event_FireGenericEvent("WindowManagementAdd", {wnd = w, strName=_tWindowManagementLabel[_ePlateType[plate.plateType]], nSaveVersion=2})
end

function TargetView:InitPromoted(plate)				-- one time window state when promoting or promoted on attach
	local w,unit = plate.nameplate, plate.unitOwner

	plate.beneBuffBar:SetUnit(unit)
	plate.harmBuffBar:SetUnit(unit)
	plate.model:SetCostume(unit)

	if not unit then
		return
	end

	-- class
	local bShow = plate.sUnitType == "Player"
	if bShow then
		local eClassId = unit:GetClassId()
		if eClassId ~= plate.eClassId then
			plate.classIcon:SetSprite(_tClassSprite[eClassId])
			plate.eClassId = eClassId
		end
	end

	if bShow ~= plate.bShowClassIcon then
		plate.classShadow:Show(bShow)
		plate.classIcon:Show(bShow)
		plate.bShowClassIcon = bShow
	end

	plate.bRequestLayout = true
end

function TargetView:AttachPlate(plate)		-- things that only change when the association changes
	local w,wf,wt,unit = plate.nameplate, plate.frame, plate.totFrame, plate.unitOwner

	-- unit
	if not unit then
		return
	end

	-- type
	local unitType = unit:GetType()
	if unitType ~= plate.sUnitType then
		plate.sUnitType = unitType
	end

	-- iconic
	local t = unit:GetArchetype()							-- very expensive
	local bIsIconic = (t and t.idArchetype or 0) == Unit.CodeEnumArchetype.Iconic
	if bIsIconic ~= plate.bIsIconic then
		plate.bIsIconic = bIsIconic
	end

	plate.bRequestRewards = true
end

function TargetView:UpdateRewards(plate)
	local ou,unit = self.userSettings, plate.unitOwner

	local GetUnitRewardIconsForm = RewardIcons and RewardIcons.GetUnitRewardIconsForm
	if not GetUnitRewardIconsForm then
		return
	end
--self._profRI = (self._profRI or 0) + 1											--PROF:

	local w = plate.rewards -- or self.wndRewardsDummy
	GetUnitRewardIconsForm(w, unit, {
		bVert = false,
	})

	local rewards = w:GetData()

	plate.bRewardIcons = rewards and rewards.nIcons and rewards.nIcons > 0


	plate.bRequestLayout = true
end

function TargetView:UpdateLayout(plate)
--Print("UpdateLayout "..plate.nAge)											--PROF:
	local o,w,wf = plate.userSettings, plate.nameplate, plate.frame


	-- clusters
	plate.clusters:ArrangeChildrenHorz(o.bFlipClusters and
		Window.CodeEnumArrangeOrigin.RightOrBottom or Window.CodeEnumArrangeOrigin.LeftOrTop)
end

function TargetView:DrawName(plate)
	local o,unit = self.userSettings, plate.unitOwner

	if _sFont ~= plate.sNameFont then
		plate.level:SetFont(_sFont)
		plate.name:SetFont(_sFont)
		plate.sNameFont = _sFont
		plate.bRequestLayout = true
	end
end

function TargetView:DrawGuild(plate)

end

function TargetView:DrawRewards(plate)

end

function TargetView:DrawLevel(plate)

end

function TargetView:DrawHealthBar(plate, bInstant)
	local o = plate.userSettings

end

function TargetView:DrawTargeting(plate, bInstant)

end


-- visibility timer		-----------------------------------------------------------------------

function TargetView:OnUpdateTimer()
	self.nAge = (self.nAge or 0) + 1
	--local nTime = GameLib.GetTickCount()												--PROF:
	--self._frameCount = (self._frameCount or 0) + 1

	self:UpdateVisibilityCommon()

--for i = 1,10 do																		--PROF:
	for _,v in next, self.arNameplates do
		self:UpdateVisibility(v)
	end
	self.bRequestRewards = false
--end																					--PROF:
	--local s2p,i2p,u2p,pool,wpool = self.arShowPlates, self.arNameplates,				--PROF:
	--self.arUnit2Plate, self.arPlatePool, self.arWindowPool							--PROF:
	--if --[[self.bIsTarget and]] self.nAge % 5 == 1 then								--PROF:
	--	Print( "" .. #i2p .. " s:" .. #s2p												--PROF:
	--		.. " p: " .. #pool .. "." .. #wpool											--PROF:
	--		.. " w: " .. (self._profWc or 0) .. "." .. (self._profWd or 0)				--PROF:
	--		.. " p: " .. (self._profPc or 0) .. "." .. (self._profPd or 0)				--PROF:
	--		.. " wc: " .. (self._profWcc or 0) .. "." .. (self._profWdc or 0))			--PROF:
	--end																				--PROF:
	--if self._frameCount % 30 == 1 then												--PROF
	--	Print((self._frameCount or 0) .. ": " .. (GameLib.GetTickCount() - nTime))		--PROF:
	--end																				--PROF:
end

function TargetView:UpdateVisibilityCommon()
	local ou = self.userSettings

	local eVisibilityToT = Apollo.GetConsoleVariable("hud.TargetOfTargetFrameDisplay")
	if eVisibilityToT ~= ou.eVisibilityToT then
		ou.eVisibilityToT = eVisibilityToT
	end

	local eTextDisplay = Apollo.GetConsoleVariable("hud.healthTextDisplay")
	if eTextDisplay ~= ou.eTextDisplay then
		ou.eTextDisplay = eTextDisplay
	end

	local nUiScale = Apollo.GetConsoleVariable("ui.Scale")

	if nUiScale ~= self.nUiScale then
		self.nUiScale = nUiScale

		local t = Apollo.GetDisplaySize()
		self.nUiWidth,self.nUiHeight = t.nRawWidth, t.nRawHeight

		self:OnSettingFontChanged()
	end
end

function TargetView:UpdateVisibility(plate)
	local o,w,unit,player = plate.userSettings, plate.nameplate, self[plate.plateType], self.unitPlayer

	if not player or not player:IsValid() then		-- world or zone unit in transition
		return
	end

	-- up-front transitions

	if unit ~= plate.unitOwner then
		plate.unitOwner = unit
		self:AttachPlate(plate)
	end

	local bIsVehicled = (unit and unit:IsInVehicle())
	if bIsVehicled ~= plate.bIsVehicled then
		plate.bIsVehicled = bIsVehicled
	end
	local unitOrVehicle = (bIsVehicled and unit:GetVehicle()) or unit
	if unitOrVehicle ~= plate.unitOrVehicle then
		if w then														-- else set by window init below if shown
			self:InitPromoted(plate)
		end
		plate.unitOrVehicle = unitOrVehicle
	end

	local eVisibility = o.cvkVisibility and Apollo.GetConsoleVariable(o.cvkVisibility)
	if eVisibility ~= o.eVisibility then
		o.eVisibility = eVisibility
	end

	-- check options

	local bShow = (unitOrVehicle and unit:IsValid()) == true
	if bShow then
		local nHealthMax = unit:GetMaxHealth() or 0
		local nHealth = nHealthMax == 0 and 0 or unit:GetHealth()

		if nHealth and nHealthMax > 0 then
			--plate:UpdatePrimaryFrame(unit)
		--elseif unit:GetName() ~= "" then
		elseif Apollo.StringLength(unit:GetName()) > 0 then		--FIXME: are they using non standard strings with 0 in them as terminator? else jsut needs == ""
		--elseif _len(unit:GetName()) > 0 then
			bShow = false
		end

		if bShow and eVisibility then				-- toggle visibility based on ui preference
			local bTestHealth

			if eVisibility == 2 then				-- always off
				bShow = false
			elseif eVisibility == 3 then			-- on only in combat
				if not player:IsInCombat() then
					bTestHealth = true
				end
			elseif eVisibility == 4 then			-- on only out of combat
				bShow = not player:IsInCombat()
			elseif eVisibility == 5 then			-- on with a target
				local target = self.unitTarget
				if not target then
					bTestHealth = true
				else
					bShow = (nHealth and nHealth ~= 0) == true
				end
			end

			if bTestHealth then
				local nShieldMax = unit:GetShieldCapacityMax() or 0
				local nShield = nShieldMax == 0 and 0 or unit:GetShieldCapacity()

				bShow = nHealth + nShield <  nHealthMax + nShieldMax
			end
		end
	end

	-- configure visibility

	--if bShow and plate.bRequestInit then												--FIXME: will be promoted on create for now
	--	self:PromotePlate(plate)
	--	w = plate.nameplate
	--end
	--if not bShow and w and (self.nAge - plate.nAge > _nReclamationAge) then			--FIXME: demotion disabled for now
	--	self:DemotePlate(plate)
	--	w = plate.nameplate
	--end
	if not w then
		return
	end

	if bShow ~= plate.bShow then
		w:Show(bShow)
		plate.bShow = bShow

		if not bShow then
			self:HideClusterFrames(plate)
		end

		local s2p = self.arShowPlates
		local count = #s2p
		if bShow then
--Print("add " .. count + 1 .. " @ " .. count .. " @ " .. #self.arNameplates)		--PROF:
			s2p[count + 1] = plate
		else
			for i = count,1,-1 do
				if s2p[i] == plate then
--Print("remove " .. i .. " @ " .. count)											--PROF:
					s2p[i] = s2p[count]; s2p[count] = nil; break
				end
			end
		end
--Print(""..#s2p)																	--PROF:
	end

	-- clusters
	for _,v in next, plate.arClusterplates do
		self:ClusterVisibility(v)
	end

	-- configure display

	if bShow then
		plate.nAge = self.nAge
		--self:DrawHealth(plate)		-- fast only

		self:DrawMisc(plate)			-- slow only.  certain death, rare, and elite

		self:DrawToT(plate)				-- slow only.  ToT update and visibility
		self:DrawClusters(plate)		-- slow only.  cluster list and visibility
		self:DrawMisc2(plate)			-- slow only.  random
		self:DrawTextColor(plate)		-- slow only.  must be after misc
		--self:DrawLevel(plate)			-- create, event

		if plate.bRequestInit then
			self:DrawName(plate)		-- create, event, target change
			--self:DrawGuild(plate)		-- create, event, target change
			plate.bRequestInit = nil
			plate.bRequestRewards = true	-- don't populate now, wait for next test. unit may be oscilating
		end
		--self:DrawRewards(plate)		-- create, slow

		if plate.bRequestLayout then
			self:UpdateLayout(plate)
			plate.bRequestLayout = false
		end
	end
end

-- configure misc items
--  sets: plate.eDisposition
function TargetView:DrawMisc(plate)
	local unit,player,unitType = plate.unitOwner, self.unitPlayer, plate.unitType

	local bDispositionChanged = false
	local eDisposition = unit:GetDispositionTo(player) or _eDisposition.Neutral		--or _eDisposition.Hostile
	if eDisposition ~= plate.eDisposition then
		plate.eDisposition = eDisposition
		bDispositionChanged = true
	end

	if self.bRequestRewards or plate.bRequestRewards then		-- needed for proper objectives check
		self:UpdateRewards(plate)								-- very expensive
		plate.bRequestRewards = nil
	end

	local bIsDead = unit:IsDead()
	if bIsDead ~= plate.bIsDead then
		plate.bIsDead = bIsDead
	end

	-- certain death

	local bShowCertainDeath = (eDisposition ~= _eDisposition.Friendly
		and unit:IsCertainDeath() and not bIsDead and unit:GetHealth())
		and true or false										-- some of these dont always return bool

	if bShowCertainDeath ~= plate.bShowCertainDeath then
		plate.certainDeath:Show(bShowCertainDeath)
		plate.bShowCertainDeath = bShowCertainDeath
	end

	-- rare and elite

	local bIsRare = unit:IsRare() == true
	if bIsRare ~= plate.bIsRare then
		plate.rare:Show(bIsRare)
		plate.bIsRare = bIsRare
	end

	local bIsElite = unit:IsElite() == true
	if bIsElite ~= plate.bIsElite then
		plate.elite:Show(bIsElite)
		plate.bIsElite = bIsElite
	end

	if bDispositionChanged then
		plate.rare:SetSprite(plate._tRareSprite[eDisposition])				--NOTE: rare and elite have no actual 'friendly' sprite,
		plate.elite:SetSprite(plate._tEliteSprite[eDisposition])			--  the carbine code bugs and always shows neutral reguardles
		plate.healthSplit:SetSprite(plate._tSplitSprite[eDisposition])
		plate.pvpMarker:SetSprite(eDisposition == _eDisposition.Hostile		--NOTE: carbine logic is unit ~= player is Enemy
			and "IconSprites:Icon_Windows16_UI_Enemy_PvPIndicator" or "IconSprites:Icon_Windows16_UI_Player_PvPIndicator")
	end

	-- group size

	local nGroupSize = (unitType == "Player" or unitType == "Pet" or unitType == "Esper Pet")
	and unit:GetGroupValue() or 0

	if nGroupSize ~= plate.nGroupSize then
		local bShow = nGroupSize > 0
		plate.groupSizeIcon:Show(bShow)
		plate.groupSizeMark:Show(bShow)

		if bShow then
			plate.groupSizeMark:SetText(nGroupSize)
			local s = BuildTooltip(String_GetWeaselString(_L.TargetFrame_Man, nGroupSize),
				String_GetWeaselString(_L.TargetFrame_GroupSize, nGroupSize))

			plate.groupSizeMark:SetTooltip(s)
			plate.groupSizeIcon:SetTooltip(s)
		end
		plate.nGroupSize = nGroupSize
	end
end

function TargetView:DrawToT(plate)				 -- called on frame
	local o,ou,unit,player = plate.userSettings, self.userSettings, plate.unitOwner, self.unitPlayer

	local bShow = o.bShowToT

	local unitToT

	if bShow then
		unitToT = unit:GetTarget()				-- this has to be polled, no event for this

		bShow = unitToT and unitToT:IsValid() and unitToT:GetHealth() and unitToT:GetMaxHealth() > 0
	end

	if bShow then
		local eVisibilityToT = ou.eVisibilityToT

		if eVisibilityToT == 2 then				-- always off
			bShow = false
		elseif eVisibilityToT == 3 then			-- on in combat
			bShow = player:IsInCombat()
		elseif eVisibilityToT == 4 then			-- on out of combat
			bShow = not player:IsInCombat()
		end
	end

	if bShow ~= plate.bShowToT then
		plate.totFrame:Show(bShow)
		plate.bShowToT = bShow
	end

	if unitToT ~= plate.unitToT then
		plate.totModel:SetCostume(unitToT)
		plate.unitToT = unitToT
	end

	if bShow then								-- this has to be polled also, pretty expensive
		local eDispositionToT = unitToT:GetDispositionTo(player) or _eDisposition.Neutral
		if eDispositionToT ~= plate.eDispositionToT then
			plate.totBacker:SetSprite(_tToTBackerSprite[eDispositionToT])
			plate.totDisposition:SetBGColor(_tToTDispositionColor[eDispositionToT])
			plate.totDisposition:SetTooltip(_tToTTooltipText[eDispositionToT])
			plate.eDispositionToT = eDispositionToT
		end
	end
end

function TargetView:DrawClusters(plate)
	local o,ou,unit,player = plate.userSettings, self.userSettings, plate.unitOwner, self.unitPlayer

	local bShow = o.bShowClusters and not plate.bIsDead

	if bShow then
		local t = ou.bShowOrganization and unit:GetClusterUnits() or {}
		local count = #t

--		if unit == player then		--NOTE: carbine had this excluded, and the the general logic inverted such that only focus target would show cluster, probably all bugs
		if ou.bShowMount and unit:IsMounted() then								-- treat mount as a cluster target
			count = count + 1
			t[count] = unit:GetUnitMount()
		end
--		end

		if ou.bShowBot and unit == player then
			if PlayerPathLib.ScientistHasScanBot() then			-- treat scanbot as a cluster target
				count = count + 1
				t[count] = PlayerPathLib.ScientistGetScanBotUnit()
			end
		end

		if ou.bShowVehicle and plate.bIsVehicled then								-- make the unit a cluster of a vehicle, if in one
			count = count + 1
			t[count] = plate.unitOrVehicle
		end

		if ou.bShowPet and unit == player then									-- treat pets as cluster targets
			local tPlayerPets = GameLib.GetPlayerPets()
			for k,v in ipairs(tPlayerPets) do
				local h = v:GetHealth() or 0
--Print("pets: " .. #tPlayerPets .. tostring(k) .. " " .. tostring(h) .. " " .. v:GetName())	--PROF:
				if k < 3 and h == 0 then
					count = count + 1
					t[count] = v
				end
			end
		end

		bShow = count > 0
		if bShow then
			for i,v in ipairs(plate.arClusterplates) do		--  need to be in order
				self:AttachCluster(v, t[i])
				if v.nameplate then
					self:InitPromotedCluster(v)
				end
			end
			--plate.bRequestLayout = true
		end
	end

	if bShow ~= plate.bShowClusters then
		if not bShow then
			self:HideClusterFrames(plate)
		end
		plate.bShowClusters = bShow
	end
end

function TargetView:DrawMisc2(plate)
	local o,unit,player,unitType = plate.userSettings, plate.unitOwner, self.unitPlayer, plate.sUnitType

	local sUnitName = unit:GetName()

	local bPlayerFlagged = player:IsPvpFlagged() and true or false
	if bPlayerFlagged ~= self.bPlayerFlagged then
		self.bPlayerFlagged = bPlayerFlagged
	end

	local bUnitFlagged = unit == player and bPlayerFlagged or unit:IsPvpFlagged()

	local sModelTooltip = ""
	if unit ~= player then
		if unitType == "Player" then
			sModelTooltip = _L.TargetFrame_IsPC
		else
			local nRank = unit:GetRank()
			if nRank ~= plate.nRank then
				plate.nRank = nRank
				plate.sRankDescription = t and 	-- weasel massively expensive
					String_GetWeaselString(_L.TargetFrame_CreatureRank, _tRankDescription[nRank])
			end
			sModelTooltip = plate.sRankDescription or ""
		end
	end
	if sModelTooltip ~= plate.sModelTooltip then
		plate.model:SetTooltip(sModelTooltip or "")
		plate.sModelTooltip = sModelTooltip
	end

	local eDisposition = plate.eDisposition

	local eBackerState = (plate.bIsDead or (unit:IsTagged() and not unit:IsTaggedByMe() and not unit:IsSoftKill()))
		and _eBackerState.Tagged or eDisposition

	if eBackerState ~= plate.eBackerState then
		plate.backer:SetSprite(plate._tBackerSprite[eBackerState])
		plate.eBackerState = eBackerState
	end

	-- attachment
	local sFaction
	if not plate.bIsDead then
		local bIsIconic = plate.bIsIconic

		if not bIsIconic and unit:GetLevel() == _nMaxLevel then
			if unitType == "Player" then
				bIsIconic = true
			elseif unitType == "Pet" then		-- all pets of iconic are iconic
				local unitOwner = unit:GetUnitOwner()
				bIsIconic = (unitOwner and sUnitName == unitOwner:GetName()) and true or false
			end
		end

		if bIsIconic then
			sFaction = _tFactionIdToType[unit:GetFaction()]	-- Dominion or Exile
		end
	end

	local eAttachment = _eAttachment.None
	if sFaction then
		eAttachment = eDisposition == _eDisposition.Friendly
			and (sFaction == "Dominion" and 2 or 3)
			or (sFaction == "Dominion" and 4 or 5)
	elseif plate.bIsVehicled then
		eAttachment = _eAttachment.Vehicle
	end

	if eAttachment ~= plate.eAttachment then
		local sAttachment = plate._tAttachmentSprite[eAttachment]
		plate.attachment:SetSprite(sAttachment)
		plate.attachment:Show(sAttachment ~= nil)
		plate.eAttachment = eAttachment
	end

	-- name and icon
	local sName,nCooldown,bForcedPvP = sUnitName, 0, false
	if bPlayerFlagged then
		local t = GameLib.GetPvpFlagInfo()
		nCooldown = ShortTime(t.nCooldown)
		bForcedPvP = t.bIsForced
	end
	if nCooldown ~= plate.nCooldownPvP then
		plate.nCooldownPvP = nCooldown
	end
	if bForcedPvP ~= plate.bForcedPvP then
		plate.bForcedPvP = bForcedPvP
	end

	if sName ~= plate.sName or bUnitFlagged ~= plate.bUnitFlagged then
		if bUnitFlagged then
			sName = String_GetWeaselString(_L.BaseBar_PvPAppend, sName)
		end
		plate.name:SetText(sName)
		plate.sName = sName
	end
	if bUnitFlagged ~= plate.bUnitFlagged then
		plate.pvpMarker:Show(bUnitFlagged)
		plate.bUnitFlagged = bUnitFlagged
	end

	local nOpacity = nCooldown > 0 and 0.5 or 1.0
	if nOpacity ~= plate.nPvpMarkerOpacity then
		plate.pvpMarker:SetOpacity(nOpacity)
		plate.nPvpMarkerOpacity = nOpacity
	end
end

-- configure text colors
function TargetView:DrawTextColor(plate)
	local unit,player = plate.unitOwner, self.unitPlayer
	local crText, crLevel

	local bIsScaled = unit:IsScaled()

	-- level

	local sLevelTooltip = (plate.sLevelTooltip or "")
	local nLevel = unit:GetLevel()

	if not nLevel then
		crLevel = _arConColor[_nConOffset]

	elseif bIsScaled then
		crLevel = _crScaling

		if unit ~= player and (nLevel ~= plate.nLevel or bIsScaled ~= plate.bIsScaled) then
			sLevelTooltip = BuildTooltip(_L.Adaptive, String_GetWeaselString(_L.TargetFrame_CreatureScales, nLevel), _csScalingHex)
		end
	else
		local n = (player:GetLevelDifferential(unit) or 0) + _nConOffset
		n = n < _nConMin and _nConMin or n > _nConMax and _nConMax or n		-- remember 'and' takes precedence
		crLevel = _arConColor[n]

		if unit ~= player and n ~= plate.nCon then
			sLevelTooltip = BuildTooltip(_arConTitle[n], String_GetWeaselString(_L.TargetFrame_TargetXPReward, _arConBody[n]), _arConTitleColor[n])
			plate.nCon = n
		end
	end

	if sLevelTooltip ~= plate.sLevelTooltip then
		plate.level:SetTooltip(sLevelTooltip)
		plate.sLevelTooltip = sLevelTooltip
	end
	if nLevel ~= plate.nLevel then
		plate.level:SetText(nLevel or "")
		plate.nLevel = nLevel
	end

	if bIsScaled ~= plate.bIsScaled then
		plate.scalingMark:Show(bIsScaled)
		plate.bIsScaled = bIsScaled
	end

	-- color and group size

	local unitType = plate.sUnitType

	if unitType == "Player" or unitType == "Pet" or unitType == "Esper Pet" then
		local owner = (unitType == "Pet" or unitType == "Esper Pet") and unit:GetUnitOwner() or unit
		local bOwnerFlag = owner:IsPvpFlagged() and true or false

		if plate.eDisposition == _eDisposition.Friendly or owner:IsThePlayer() then
			crText = bOwnerFlag and _crAlly
				or owner:IsInYourGroup() and _crGroup
				or _crAllyUnflag
		else
			local bPlayerFlagged = plate.bPlayerFlagged
			crText = bOwnerFlag ~= bPlayerFlagged and _crEnemyMismatch
				or bOwnerFlag and _crEnemy or _crEnemyUnflag
		end

	else													-- npc
		crText = _tTextColors[plate.eDisposition]
	end

	if crText ~= plate.crText then
		plate.name:SetTextColor(crText)
		plate.crText = crText
	end
	if crLevel ~= plate.crLevel then
		plate.level:SetTextColor(crLevel)
		plate.crLevel = crLevel
	end
end


-- other system events	-----------------------------------------------------------------------

function TargetView:OnRequestRewardsUpdate()
--self._profRUA = (self._profRUA or 0) + 1									--PROF:
	self.bRequestRewards = true
end

function TargetView:OnCharacterLoaded()
--self._profE = (self._profE or 0) + 1										--PROF:
	local u2p,unit = self.arUnit2Plate, GameLib.GetPlayerUnit()
	local plate = u2p.unitPlayer

	self.unitPlayer = unit and unit:IsValid() and unit or nil

	if plate and plate.nameplate then
		self:UpdateVisibility(plate)
	end
end

function TargetView:OnTargetUnitChanged(unit)
--self._profE = (self._profE or 0) + 1										--PROF:
	local u2p = self.arUnit2Plate
	local plate = u2p.unitTarget

	self.unitTarget = unit and unit:IsValid() and unit or nil

	if plate and plate.nameplate then
		self:UpdateVisibility(plate)
	end
end

function TargetView:OnAlternateTargetUnitChanged(unit)
--self._profE = (self._profE or 0) + 1										--PROF:
	local u2p = self.arUnit2Plate
	local plate = u2p.unitFocus

	self.unitFocus = unit and unit:IsValid() and unit or nil

	if plate and plate.nameplate then
		self:UpdateVisibility(plate)
	end
end

function TargetView:OnFocusSlashCommand()
	self.unitFocus:SetAlternateTarget(GameLib.GetTargetUnit())
end

function TargetView:HideClusterFrames(plate)
	for _,v in next, plate.arClusterplates do
		self:AttachCluster(v, nil)
		self:ClusterVisibility(v)
	end
end

function TargetView:OnResolutionChanged()
	self.nUiScale = nil													-- force scale re-evaluation
	self:OnUpdateTimer()												-- immediate, so no post rez change pop
end


-----------------------------------------------------------------------------------------------
-- ClusterFrame Functions
-----------------------------------------------------------------------------------------------

function TargetView:DrawHealthCluster(cluster)
	local w,unit,player = cluster.nameplate, cluster.unitOwner, self.unitPlayer
	if not unit or not cluster.healthFill then		--FIXME: after the show functionality is complete, remove
		return
	end

	local nHealthMax = unit:GetMaxHealth()
	local nHealth    = nHealthMax == 0 and 0 or unit:GetHealth()
	local nShieldMax = unit:GetShieldCapacityMax() or 0
	local nShield    = nShieldMax == 0 and 0 or unit:GetShieldCapacity()

	local bNewHealth = nHealth ~= cluster.nHealth or nHealthMax ~= cluster.nHealthMax
	local bNewShield = nShield ~= cluster.nShield or nShieldMax ~= cluster.nShieldMax

	local bShowHealth = nHealthMax and nHealthMax > 0
	local bShowShield = nShieldMax > 0
	local bNewShowShield = bShowShield ~= cluster.bShowShield

	if bNewShowShield or bShowHealth ~= cluster.bShowHealth then
		cluster.healthFill:Show(not bShowShield and bShowHealth)
		cluster.healthShieldFill:Show(bShowShield)
		cluster.shieldFill:Show(bShowShield)
		cluster.cover:SetSprite(bShowHealth and (bShowShield
			and "spr_TargetFrame_ClusterCoverShield" or "spr_TargetFrame_ClusterCover") or nil)
		cluster.bShowHealth, cluster.bShowShield = bShowHealth, bShowShield
	end

	local ratioH
	if bShowHealth and (bNewHealth or bNewShowShield) then
		local wh = bShowShield and cluster.healthShieldFill or cluster.healthFill
		ratioH = nHealth / (nHealthMax or nHealth)

		if nHealthMax ~= cluster.nHealthMax or bNewShowShield then
			wh:SetMax(nHealthMax)
			cluster.nHealthMax = nHealthMax
		end
		if nHealth ~= cluster.nHealth or bNewShowShield then
			wh:SetProgress(nHealth)
			cluster.nHealth = nHealth
		end

		local eTintType = unit:IsInCCState(_eStateVulnerability) and _eTintType.Vulnerable
			or (ratioH <= _nHealthRed and _eTintType.Critical)
			or (ratioH <= _nHealthYellow and _eTintType.Warning)
			or _eTintType.Good

		if eTintType ~= cluster.eHealthTint or bNewShowShield then
			wh:SetFullSprite(_tClusterBar[eTintType])
			cluster.eHealthTint = eTintType
		end
	end

	if nShieldMax ~= cluster.nShieldMax then
		cluster.shieldFill:SetMax(nShieldMax)
		cluster.nShieldMax = nShieldMax
	end

	if nShield ~= cluster.nShield then
		cluster.shieldFill:SetProgress(nShield)
		cluster.nShield = nShield
	end

	-- tooltip text
	local f = String_GetWeaselString

	if bNewHealth and bShowHealth then
		local n,m,r = ShortNumber(nHealth), ShortNumber(nHealthMax), _floor(ratioH * 100)
		bNewHealth = n ~= cluster.nHealthClip or m ~= cluster.nHealthMaxClip or r ~= cluster.nHealthRatio
		if bNewHealth then
			cluster.sHealthTooltip = f(_L.TargetFrame_HealthFormat, _L.Innate_Health,					-- massively expensive
				FormatNumber(n), FormatNumber(m), f(_L.CRB_Percent, r))
			cluster.nHealthClip, cluster.nHealthMaxClip, cluster.nHealthRatio = n,m,r
		end
	end

	if bNewShield and bShowShield then
		local n,m,r = ShortNumber(nShield), ShortNumber(nShieldMax), _floor(nShield / nShieldMax * 100)
		bNewShield = n ~= cluster.nShieldClip or m ~= cluster.nShieldMaxClip or r ~= cluster.nShieldRatio
		if bNewShield then
			cluster.sShieldTooltip = f(_L.TargetFrame_HealthFormat, _L.Character_ShieldLabel,
				FormatNumber(n), FormatNumber(m), f(_L.CRB_Percent, r))
			cluster.nShieldClip, cluster.nShieldMaxClip, cluster.nShieldRatio = n,m,r
		end
	end

	local sUnitName = unit:GetName()
	if bNewHealth or bNewShield or sUnitName ~= cluster.sUnitName then
		local s = sUnitName
		if bShowHealth then
			s = s .. "\n" .. cluster.sHealthTooltip
		end
		if bShowShield then								-- 100% non-combat, 95%+ combat
			s = s .. "\n" .. cluster.sShieldTooltip
		end
		w:SetTooltip(s)
		cluster.sUnitName = sUnitName
	end
end

function TargetView:DrawCastCluster(cluster)
	local unit = cluster.unitOwner
	if not unit or not cluster.healthFill then		--FIXME: after the show functionality is complete, remove
		return
	end

	local bShowCasting, eType = (unit:ShouldShowCastBar())
	if bShowCasting then
		eType = unit:GetCastBarType()
		bShowCasting = (eType and eType ~= _eCastType.None) == true
	end

	if bShowCasting ~= cluster.bShowCasting then
		cluster.castBar:Show(bShowCasting)
		cluster.bShowCasting = bShowCasting
	end

	if bShowCasting then
		--local sSpellName = unit:GetCastName()
		local sIconSprite = unit:GetCastDuration() > 0
			and (eType == _eCastType.Telegraph_Evade and "CRB_TargetFrameSprites:sprTF_CastIconEvade"
			or eType == _eCastType.Telegraph_Backlash and "CRB_TargetFrameSprites:sprTF_CastIconInterrupt")
			or ""

		if sIconSprite ~= cluster.sIconSprite then
			cluster.castIcon:SetSprite(sIconSprite)
			cluster.sIconSprite = sIconSprite
		end
	end
end


-- create destroy		-----------------------------------------------------------------------

function TargetView:OnClusterCreated(plate)
--self._profE = (self._profE or 0) + 1												--PROF:
--local unitType = unit:GetType()
--Print("create: " .. tostring(unit:GetName()))										--PROF:
--	.. " age:" .. plate.nAge .. " now:" .. self.nAge)								--PROF:
	local i2c,pool = plate.arClusterplates, self.arPlatePool

	--local count = #pool
	--if count > 0 then												-- check plate pool
	--	plate, pool[count] = pool[count], nil
	--else
		cluster = { }
--self._profPc = (self._profPc or 0) + 1											--PROF:
	--end

	-- set plate

	self:InitCluster(cluster, plate, unit, unitType)

	i2c[#i2c + 1] = cluster

	cluster.bRequestInit = true			-- may never happen
end

function TargetView:InitCluster(cluster, plate, unit, unitType)
	if not self.tClusterTemplate then	-- make static template with base values, to know what to clean or migrate
		self.tClusterTemplate = { }
		self:InitCluster(self.tClusterTemplate, "unitPlayer", unit, "Player")
--Print("----")																		--PROF:
--for k,v in next, self.tClusterTemplate do											--PROF:
--	Print(" " .. k .. ": " .. tostring(v))											--PROF:
--end																				--PROF:
	end

	cluster.tPlate 				= plate
	cluster.unitOwner			= unit
	cluster.sUnitType			= unitType

	cluster.nAge				= 0
end

function TargetView:InitPromotedCluster(cluster)				-- one time window state when promoting or promoted on attach
	local w,unit = cluster.nameplate, cluster.unitOwner

	cluster.model:SetCostume(unit)

	if not unit then
		return
	end
end

function TargetView:DemoteCluster(cluster)
	local w = cluster.nameplate
	if not w then
		return
	end

--Print("demote " .. (cluster.plateType or "unknown") .. " " .. (cluster.sUnitName or "unknown"))	--PROF:

	local wpool,ctemplate = self.arWindowPool, self.tClusterTemplate
	local wcount = #wpool

	w:SetData(nil)
	--plate.beneBuffBar:SetUnit(nil)
	--plate.harmBuffBar:SetUnit(nil)
	cluster.model:SetCostume(nil)

	for k,_ in next, cluster do
		if ctemplate[k] == nil then cluster[k] = nil end
	end
	cluster.nameplate = nil						-- just in case
	w:Destroy()
--self._profWdc = (self._profWdc or 0) + 1																--PROF:

	cluster.bRequestInit = true
end

function TargetView:PromoteCluster(cluster)
	local plate = cluster.tPlate
	local o,w,pool = plate.userSettings, cluster.nameplate, self.arWindowPool

	if not w or not w:IsValid() then
		w = Apollo.LoadForm(self.xmlDoc, "Mini", plate.clusters, self)
--self._profWcc = (self._profWcc or 0) + 1																--PROF:
		cluster.nameplate = w
		cluster.model = w:FindChild("Model")
		cluster.rewards = w:FindChild("GoalPanel")
		cluster.level = w:FindChild("Level")
		cluster.scalingMark = w:FindChild("ScalingMark")
		cluster.certainDeath = w:FindChild("CertainDeath")
		cluster.cover = w:FindChild("Cover")
		cluster.healthFill = w:FindChild("HealthFill")
		cluster.healthShieldFill = w:FindChild("HealthShieldFill")
		cluster.shieldFill = w:FindChild("ShieldFill")
		cluster.castBar = w:FindChild("CastBar")
		cluster.castIcon = w:FindChild("CastIcon")
	end

	-- set window
	--plate.bShow = nil
	w:SetData(cluster)
	--w:SetUnit(plate.unitOrMount, 1)
	self:InitPromotedCluster(cluster)

	-- one time fix-ups

end

function TargetView:AttachCluster(cluster, unit)
	local w = cluster.nameplate

	if unit and unit == cluster.unitOwner then
		return
	end

	cluster.unitOwner = unit

	cluster.bRequestRewards = true
end


-- visibility timer		-----------------------------------------------------------------------

function TargetView:ClusterVisibility(cluster)
	local plate = cluster.tPlate
	local w,unit = cluster.nameplate, cluster.unitOwner

	if not cluster.nAge then		--NOTE: carbine race condition, via Lua code this should never happen
		return
	end

	local bShow = (plate and unit and unit:IsValid() and not unit:IsDead()) and true or false

	-- configure visibility

	if bShow and cluster.bRequestInit then
		self:PromoteCluster(cluster)
		w = cluster.nameplate
	end
	if not bShow and w and (self.nAge - cluster.nAge > _nReclamationAge) then
		self:DemoteCluster(cluster)
		w = cluster.nameplate
	end
	if not w then
		return
	end

	if bShow ~= cluster.bShow then
		w:Show(bShow)
		cluster.bShow = bShow

		local s2c = plate.arShowClusters
		local count = #s2c
		if bShow then
--Print("add " .. count + 1 .. " @ " .. count .. " @ " .. #plate.arClusterplates)	--PROF:
			s2c[count + 1] = cluster
		else
			for i = count,1,-1 do
				if s2c[i] == cluster then
--Print("remove " .. i .. " @ " .. count)											--PROF:
					s2c[i] = s2c[count]; s2c[count] = nil; break
				end
			end
		end
--Print("c: "..#s2c)																--PROF:
		plate.bRequestLayout = true				-- number shown changed, re-layout
	end

	-- configure display

	if bShow then
		cluster.nAge = self.nAge

		self:DrawMiscCluster(cluster)	-- slow
		self:DrawTextColorCluster(cluster)	-- slow

		if cluster.bRequestInit then
			--self:DrawLevel(plate)		-- create, event
			--self:DrawName(plate)		-- create, event, target change
			--self:DrawGuild(plate)		-- create, event, target change
			cluster.bRequestInit = nil
			cluster.bRequestRewards = true	-- don't populate now, wait for next test. unit may be oscilating
		end
		--self:DrawRewards(plate)		-- create, slow
	end
end

function TargetView:DrawMiscCluster(cluster)
	local w,unit = cluster.nameplate, cluster.unitOwner

	if self.bRequestRewards or cluster.bRequestRewards then		-- needed for proper objectives check
		self:UpdateRewards(cluster)									-- very expensive
		cluster.bRequestRewards = nil
	end

	local bShowRewards = cluster.bRewardIcons
	if bShowRewards ~= cluster.bShowRewards then
		cluster.rewards:IsShown(bShowRewards, true)
		cluster.bShowRewards = bShowRewards
	end
end

function TargetView:DrawTextColorCluster(cluster)
	local unit,player = cluster.unitOwner, self.unitPlayer

	local bIsScaled = unit:IsScaled()
	local nLevel = unit:GetLevel()

	local sLevelTooltip, crLevel = (cluster.sLevelTooltip or "")

	if not nLevel then
		crLevel = _arConColor[_nConOffset]

	elseif bIsScaled then
		crLevel = _crScaling

		if unit ~= player  and (nLevel ~= cluster.nLevel or bIsScaled ~= cluster.bIsScaled) then
			sLevelTooltip = BuildTooltip(_L.Adaptive, String_GetWeaselString(_L.TargetFrame_CreatureScales, nLevel), _csScalingHex)
		end
	else
		local n = (player:GetLevelDifferential(unit) or 0) + _nConOffset
		n = n < _nConMin and _nConMin or n > _nConMax and _nConMax or n
		crLevel = _arConColor[n]

		if unit ~= player and n ~= cluster.nCon then
			sLevelTooltip = BuildTooltip(_arConTitle[n], String_GetWeaselString(_L.TargetFrame_TargetXPReward, _arConBody[n]), _arConTitleColor[n])
			cluster.nCon = n
		end
	end

	if sLevelTooltip ~= cluster.sLevelTooltip then
		cluster.level:SetTooltip(sLevelTooltip)
		cluster.sLevelTooltip = sLevelTooltip
	end
	if crLevel ~= cluster.crLevel then
		cluster.level:SetTextColor(crLevel)
		cluster.crLevel = crLevel
	end
	local sLevel = nLevel or ""--cluster.sUnitName
	if sLevel ~= cluster.sLevel then
		cluster.level:SetText(sLevel)
		cluster.nLevel = nLevel
		cluster.sLevel = sLevel
	end

	if bIsScaled ~= cluster.bIsScaled then
		cluster.scalingMark:Show(bIsScaled)
		cluster.bIsScaled = bIsScaled
	end
end


-- other system events	-----------------------------------------------------------------------

---------------------------------------------------------------------------------------------------
-- Tutorial anchor request
---------------------------------------------------------------------------------------------------

function TargetView:OnTutorial_RequestUIAnchor(eAnchor, idTutorial, strPopupText)
	local u2p = self.arUnit2Plate

	local tAnchors = {
		[GameLib.CodeEnumTutorialAnchor.HealthBar] 		= true,
		[GameLib.CodeEnumTutorialAnchor.InterruptArmor] = true,
		[GameLib.CodeEnumTutorialAnchor.BuffFrame] 		= true,
		[GameLib.CodeEnumTutorialAnchor.ShieldBar]		= true,
	}

	if not tAnchors[eAnchor] then
		return
	end

	local tAnchorMapping = {
		[GameLib.CodeEnumTutorialAnchor.HealthBar] 		= u2p.unitPlayer.health,
		[GameLib.CodeEnumTutorialAnchor.InterruptArmor] = u2p.unitTarget.ccArmor,
		[GameLib.CodeEnumTutorialAnchor.BuffFrame] 		= u2p.unitPlayer.frame:FindChild("BuffTutorialAnchor"),
		[GameLib.CodeEnumTutorialAnchor.ShieldBar]		= u2p.unitPlayer.shield
	}

	if tAnchorMapping[eAnchor] then
		Event_FireGenericEvent("Tutorial_ShowCallout", eAnchor, idTutorial, strPopupText, tAnchorMapping[eAnchor])
	end
end


-----------------------------------------------------------------------------------------------
-- Nameplate Form Functions
-----------------------------------------------------------------------------------------------

function TargetView:OnGenerateBuffTooltip(wndHandler, wndControl, eType, splBuff)
	if wndHandler == wndControl or eType ~= Tooltip.TooltipGenerateType_Spell then
		return	-- return if the control is the container, only the unnamed buf control shall pass
	end

	local w,plate = (wndHandler)
	while w and not plate do										-- find root window and plate
		plate = w:GetData()
		w = not plate and w:GetParent() or w						-- skip expensive call
	end

	if plate then
		Tooltip.GetBuffTooltipForm(plate, wndControl, splBuff, { bFutureSpell = false })
	end
end

function TargetView:OnMouseButtonDown(wndHandler, wndControl, eMouseButton, x, y)
	local ou,w,plate = self.userSettings, (wndHandler == wndControl and wndHandler or nil)

	while w and not plate do										-- find root window and plate
		plate = w:GetData()
		w = not plate and w:GetParent() or w						-- skip expensive call
	end

	local target = plate											-- be strict, control must be handler, handlers must have valid plates.
		and ((wndHandler == plate.totFrame and plate.unitToT)		--   this was previously firing up to 7-8 times per-click on any part
		or (wndHandler == plate.model and plate.unitOwner))			--   of the frame, adding to targeting issues, random UI DoS, and lag.

	if target then
		if eMouseButton == GameLib.CodeEnumInputMouse.Left then
			GameLib.SetTargetUnit(target)
			return false
		end

		if eMouseButton == GameLib.CodeEnumInputMouse.Right then
			Event_FireGenericEvent("GenericEvent_NewContextMenuPlayerDetailed", nil, target:GetName(), target)
			return true
		end
	end

	if ou.bIsDemo then								-- IsDemo()
		return true
	end

	return false
end

function TargetView:OnQueryDragDrop(wndHandler, wndControl, nX, nY, wndSource, strType, nValue)
	if wndHandler ~= wndControl then
		return Apollo.DragDropQueryResult.PassOn
	end

	local unitToT = GameLib.GetTargetUnit()		--FIXME: never seemd to work on drop
	if unitToT == nil then
		return Apollo.DragDropQueryResult.Invalid
	end

	if strType == "DDBagItem" and unitToT:IsACharacter() and not unitToT:IsThePlayer() then
		return Apollo.DragDropQueryResult.Accept
	end

	return Apollo.DragDropQueryResult.Invalid
end

function TargetView:OnDragDrop(wndHandler, wndControl, nX, nY, wndSource, strType, nValue)
	if wndHandler ~= wndControl then
		return false
	end

	local unitToT = GameLib.GetTargetUnit()
	if unitToT == nil then
		return false
	end
	if strType == "DDBagItem" and unitToT:IsACharacter() and not unitToT:IsThePlayer() then
		Event_FireGenericEvent("ItemDropOnTarget", unit, strType, nValue)
		return false
	end
end

function TargetView:OnGeneratePvPTooltip(wndHandler, wndControl)
	if wndHandler ~= wndControl then
		return
	end

	local w,plate = (wndHandler)
	while w and not plate do										-- find root window and plate
		plate = w:GetData()
		w = not plate and w:GetParent() or w						-- skip expensive call
	end

	local unit = plate and plate.unitOwner
	if not unit or not plate.bUnitFlagged or not unit:IsValid() then
		return
	end

	local s = _L.TargetFrame_PvPMarked

	if unit == self.unitPlayer then
		s = '<T TextColor="UI_WindowTextDefault">' .. _L.TargetFrame_PvPBonus .. '</T>'
		if plate.nCooldownPvP > 0 then
			s = s .. '<P></P><T TextColor="UI_WindowTextDefault">' ..
				String_GetWeaselString(_L.TargetFrame_PvPCooldown, FormatTime(plate.nCooldownPvP)) .. '</T>'
		else
			s = s .. '<P></P><T TextColor="Disabled">' .. (plate.nForcedPvP
				and _L.TargetFrame_PvPForced or _L.TargetFrame_PvPOff) .. '</T>'
		end
	end

	if s ~= plate.sTooltipPvP then
		wndControl:SetTooltip(s)
		plate.sTooltipPvP = s
	end
end

--function TargetView:OnMouseButtonDownPvP(wndHandler, wndControl, eMouseButton, x, y) --FIXME: is carbine logic, im not sure was really intended to affect ToT, doesnt work well
--	local ou,w,plate = self.userSettings, (wndHandler == wndControl and wndHandler or nil)

--	while w and not plate do										-- find root window and plate
--		plate = w:GetData()
--		w = not plate and w:GetParent() or w						-- skip expensive call
--	end

--	local unitToT = plate and plate.unitToT
--	if unitToT then
--		if eMouseButton == GameLib.CodeEnumInputMouse.Left then
--			GameLib.SetTargetUnit(unitToT)
--			return false
--		end
--		if eMouseButton == GameLib.CodeEnumInputMouse.Right then
--			Event_FireGenericEvent("GenericEvent_NewContextMenuPlayerDetailed", nil, unitToT:GetName(), unitToT)
--			return true
--		end
--	end

--	if ou.bIsDemo then								-- IsDemo()
--		return true
--	end

--	return false
--end


-----------------------------------------------------------------------------------------------
-- Interface AddOn Configuration Functions
-----------------------------------------------------------------------------------------------

function TargetView:OnUnitFrameOptionsUpdated()
	local i2p = self.arNameplates

	-- clear
	while #i2p > 0 do
		self:OnUnitDestroyed(i2p[1])
	end

	-- build
	for _,v in next, { "unitPlayer", "unitTarget", "unitFocus" } do
		self:OnUnitCreated(v)
	end
end

function TargetView:OnKeyBindingUpdated(strKeybind)				-- results of function not used
	if strKeybind ~= "Path Action" and strKeybind ~= "Cast Objective Ability" then
		return
	end

	local bind = GameLib.GetKeyBinding("PathAction")
	self.bPathActionUsesIcon = bind == "Unbound" or #bind > 1			-- dont show interact

	bind = GameLib.GetKeyBinding("CastObjectiveAbility")
	self.bQuestActionUsesIcon = bind == "Unbound" or #bind > 1			-- dont show interact
end


-----------------------------------------------------------------------------------------------
-- OptionsForm Form Functions
-----------------------------------------------------------------------------------------------

function TargetView:OnConfigure()
	self:OnOptionsForm()
end

function TargetView:OnCloseWindow()
	self.wndOpts:Close()
end

function TargetView:OnCancel()
	self:OnCloseWindow()
end

function TargetView:OnOK()
	self.configCommitting = true
	self:OnCloseWindow()
end

function TargetView:OnOptionsClosed(wndHandler, wndControl)
	if wndHandler ~= wndControl then
		return
	end
	local wo,L = self.wndOpts, self.L
	if wo and wo:IsValid() then
		wo:Close()
		wo:Destroy()
		self.wndOpts = nil
		self._tOptionsTabs = nil
		self.woChoiceScroll = nil

		Event_FireGenericEvent("WindowManagementRemove", {strName=L["TargetView Options"]})
	end
end

function TargetView:OnOptionsForm()
	local L = self.L

	local wo = Apollo.LoadForm(self.xmlDoc, "OptionsForm", nil, self)
	self.wndOpts = wo

	Apollo.LoadForm(self.xmlDoc, "NormalView", wo:FindChild("NormalContent"), self)
	Apollo.LoadForm(self.xmlDoc, "TargetView", wo:FindChild("TargetContent"), self)
	Apollo.LoadForm(self.xmlDoc, "AdvancedView", wo:FindChild("AdvancedContent"), self)

	self:OnPanelForm(wo)

	-- fixups
	self._tOptionsTabs = {
		NormalViewButton = "NormalContent",
		TargetViewButton = "TargetContent",
		AdvancedViewButton = "AdvancedContent"
	}
	local w = wo:FindChild(self.sOptionsTab or "NormalViewButton")
	self:OnOptionsTabCheck(nil, w)
	w:SetCheck(true)
	wo:FindChild("TargetViewButton"):Enable(false)			-- no other pages for now
	wo:FindChild("AdvancedViewButton"):Enable(false)

	self:SetButtonEnabled(wo:FindChild("MainShowGroup"), false, true)	-- not doing these for now
	--wo:FindChild("MainShowGroup"):Enable(false)

	wo:FindChild("Title"):SetText(self.L["TargetView Options"])

	self:OnCurrentToConfig()
	wo:Invoke()

	Event_FireGenericEvent("WindowManagementAdd", {wnd = wo, strName=L["TargetView Options"], nSaveVersion=_VersionWindow})
end

function TargetView:OnCurrentToConfig()
	local o,wo = self.userSettings, self.wndOpts

	-- Generic managed controls
	self:OnCurrentToPanel(wo)

	wo:FindChild("ResetDropToggle"):SetText(o.sSettingsName)

	self:OnEnablerButtonChecked()												-- enable disable buttons
end


function TargetView:OnEnablerButtonChecked()
	local o,wo = self.userSettings, self.wndOpts

	self:SetButtonEnabled(wo:FindChild("MainShowClusterMine"), o.bShowClusters)
	self:SetButtonEnabled(wo:FindChild("MainShowClusterTarget"), o.bShowClusters)
	self:SetButtonEnabled(wo:FindChild("MainShowClusterFocus"), o.bShowClusters)
end

function TargetView:OnSettingFontChanged()
	local o,wo = self.userSettings, self.wndOpts

--	_crGuild = o.bColorLightGuild and _crGuild_Light or _crGuild_Dark
	_crEnemy = o.bColorLightEnemy and _crEnemy_Light or _crEnemy_Dark
	_tTextColors[_eDisposition.Hostile] = _crEnemy

	local n = self.nUiScale

	n = (n < 0.79 and (self.nUiWidth < 1030 or self.nUiHeight < 782) and 0.65)	-- panic modes for threshholds when the carbine client becomes unreadable
		or (n < 1.1 and (self.nUiWidth < 1020 or self.nUiHeight < 600) and 0.65) or n

	_sFont = not o.bFontLargeName and _sFont_Default
		or (n > 0.86 and (o.bFontLargerName and _sFont_Larger or _sFont_Large))
		or (n > 0.78 and (o.bFontLargerName and _sFont_Larger85 or _sFont_Large85))
		or (n >= 0.70 and (o.bFontLargerName and _sFont_Larger75 or _sFont_Large75))
		or (o.bFontLargerName and _sFont_Larger70 or _sFont_Large70)

--Print(_sFont .. " " .. --[[ _nNearClip .. " " ..]] self.nUiWidth .. " " .. self.nUiHeight)									--PROF:

	for _,v in next, self.arNameplates do
		v.crText = nil
		--v.crLevel = nil
		if v.nameplate and v.unitOwner then
			self:DrawTextColor(v)
			self:DrawName(v)
		end
	end

	if wo then
		self:OnEnablerButtonChecked()
	end
end

function TargetView:OnSettingClusterOption()		-- cluster changes often require layout, so skip delay layout 'pop'
	for _,v in next, self.arNameplates do
		if v.nameplate then
			self:UpdateVisibility(v)
		end
	end
	self:OnEnablerButtonChecked()
end

function TargetView:OnSettingRebuildFrames()
	self:OnUnitFrameOptionsUpdated()
	for _,v in next, self.arNameplates do
		if v.nameplate then
			self:UpdateVisibility(v)
		end
	end
	self:OnEnablerButtonChecked()
end


-----------------------------------------------------------------------------------------------
-- TargetView Startup
-----------------------------------------------------------------------------------------------

function TargetView:OnSave(eType)
	local tSave

	if eType == GameLib.CodeEnumAddonSaveLevel.Character then
		local o = self.userSettings

		tSave = { }
	end

	if eType == GameLib.CodeEnumAddonSaveLevel.Realm then
	end

	if tSave then
		local o = self.userSettings

		for k,v in pairs(o) do
			local t = _defaultSettings[k]
			if (type(t) == "table" and t.eSaveLevel or _eSaveLevel.Character) == eType then
				tSave[k] = DeepCopy(v)
			end
		end
	end

	return tSave
end

function TargetView:OnRestore(eType, tSave)
	if tSave then
		local o,_ds = self.userSettings, _defaultSettings

		local nVersion = tSave["nSettingsVersion"] or 0

		for k,v in pairs(tSave) do
			local t = _defaultSettings[k]
			if (type(t) == "table" and t.eSaveLevel or _eSaveLevel.Character) == eType then
				-- migrations
				-- filters
				if k == "nSettingsVersion" then v = o.nSettingsVersion end
				if k == "sSettingsTooltip" then v = nil end
				o[k] = DeepCopy(v)
			end
			-- migrations
		end
	end

	if eType == GameLib.CodeEnumAddonSaveLevel.Character then
		self:InitializeWindows()
	end
end

function TargetView:InitializeWindows()
	local o = self.userSettings

	self:OnCurrentToCommand()

	self.nUiScale = self.nUiScale or Apollo.GetConsoleVariable("ui.Scale")
	local t = Apollo.GetDisplaySize()
	self.nUiWidth,self.nUiHeight = t.nRawWidth, t.nRawHeight

	-- re-bind any newly loaded custom reset settings to base custom settings
	self._resetSettingsCustom = self._resetSettingsCustom or _resetSettings.custom
	_resetSettings.custom = setmetatable(o.tSettingsCustom, { __index = self._resetSettingsCustom })

	-- rest needs docload
	if not self.onXmlDocLoadedCalled then
		return
	end

	o.bIsDemo = IsDemo()						-- too expensive to call 8 times per mousedown event

	local player = GameLib.GetPlayerUnit()
	self.unitPlayer = player

	-- loaded units
	self:OnSettingFontChanged()					-- update any existing units, includes OnSettingNameChanged(), OnSettingFontChanged()
	--self:OnSettingHealthChanged()

	self:OnKeyBindingUpdated("Path Action")		-- let binding dependent settings initialize
end


-----------------------------------------------------------------------------------------------
-- TargetView Commands
-----------------------------------------------------------------------------------------------

function TargetView:OnShowMenu()			-- used by InterfaceMenuList
	self:OnCommand()
end

function TargetView:OnCurrentToCommand()
	local o = self.userSettings
	print = o.bIsDebug and _print or _dummy
end

function TargetView:OnCommand(sCommand, sArgs)
	local o,wo,L = self.userSettings, self.wndOpts, self.L

	sArgs = sArgs and sArgs:lower() or ""
	local sArg1, sArg2 = sArgs:match("([^ ]+) *(.-) *$")	-- '  aaa   bb cc  ' to 'aaa', 'bb cc'

	if not sArg1 or sArg1 == ""
		or sArg1 == "options"								-- toggle config window
		or sArg1 == "config" or sArg1 == "settings" then
		if wo then
			self:OnOptionsClosed()
		else
			self:OnOptionsForm()
		end
	elseif sArg1 == "optionsopen" then
		if not wo then
			self:OnOptionsForm()
		end
	elseif sArg1 == "optionsclose" then
		if wo then
			self:OnOptionsClosed()
		end

	elseif sArg1 == "reset" then
		local t = (not sArg2 or sArg2 == "") and _resetSettings.default or _resetSettings[sArg2]
		if not t then
			PostChat(_format(L["No reset settings named '%s' for %s."], sArg2, _AddonName))
			return
		end
		o = SettingsCopy(_defaultSettings); self.userSettings = o
		o.tSettingsCustom = _resetSettings.custom
		t.nSettingsVersion = t.nSettingsVersion or o.nSettingsVersion
		self:OnRestore(GameLib.CodeEnumAddonSaveLevel.Realm, t)			-- realm first because char fires init
		self:OnRestore(GameLib.CodeEnumAddonSaveLevel.Character, t)
		if wo and wo:IsValid() then
			self:OnCurrentToConfig()
		end

	elseif sArg1 == "debug" then
		o.bIsDebug = not o.bIsDebug
		self:OnCurrentToCommand()

	elseif sArg1 == "help" then
		local _addonname = _AddonName:lower()
		PostChat(_format(L["%s commands:"],_AddonName))
		PostChat(_format(L[" /%s - show options"],_addonname))
		PostChat(_format(L[" /%s reset - reset to defaults"],_addonname))
		PostChat(_format(L[" /%s reset [name] - reset to named preset"],_addonname))
		PostChat(_format(L[" /%s help - help message"],_addonname))
	end
end


-----------------------------------------------------------------------------------------------
-- FrameView Instance
-----------------------------------------------------------------------------------------------

local TargetViewInst = TargetView:new()
TargetViewInst:Init()

